from typing import List

from neurosym.types.type_with_environment import Environment, TypeWithEnvironment

from .preorder_mask import PreorderMask


class TypePreorderMask(PreorderMask):
    """
    Masks out productions that would lead to ill-typed programs.
    """

    def __init__(self, tree_dist, dsl):
        super().__init__(tree_dist)
        self.dsl = dsl
        assert len(self.dsl.valid_root_types) == 1, "Only one root type is supported"
        self.root_type = self.dsl.valid_root_types[0]

        # stack of list of type_with_environment objects
        # each list represents the types of the children of the current node
        self.type_stack: List[List[TypeWithEnvironment]] = []

    def compute_mask(self, position, symbols):
        valid_productions = {
            self.tree_dist.symbol_to_index[sym.symbol()]
            for sym, _ in self.dsl.productions_for_type(self.type_stack[-1][position])
        }
        return [i in valid_productions for i in symbols]

    def on_entry(self, position, symbol):
        symbol, arity = self.tree_dist.symbols[symbol]
        if symbol == "<root>":
            self.type_stack.append(
                [TypeWithEnvironment(self.root_type, Environment.empty())]
            )
            return
        parent_type = self.type_stack[-1][position]
        production = self.dsl.get_production(symbol)
        children_types = production.type_signature().unify_return(parent_type)
        if children_types is None:
            raise ValueError(
                f"Type mismatch in production {production} with parent type {parent_type}"
            )
        assert len(children_types) == arity
        self.type_stack.append(children_types)

    def on_exit(self, position, symbol):
        del position
        del symbol
        self.type_stack.pop()



import queue
from dataclasses import dataclass, field
from typing import Callable

from neurosym.programs.s_expression import SExpression
from neurosym.search_graph.search_graph import SearchGraph


def bounded_astar(
    g: SearchGraph, cost_plus_heuristic: Callable[[SExpression], float], max_depth: int
):
    """
    Performs a bounded a-star search on the given search graph, yielding each node in
    the order it was visited.

    :param g: Search graph to search over
    :param cost_plus_heuristic: Cost plus heuristic function to use for A*.
        The heuristic function should be admissible, i.e. it should never overestimate
        the cost to reach the goal.
    :param max_depth: Maximum depth to search to.
    """
    assert max_depth > 0
    visited = set()
    fringe = queue.PriorityQueue()

    def add_to_fringe(node, depth):
        fringe.put(BoundedAStarNode(cost_plus_heuristic(node), node, depth))

    add_to_fringe(g.initial_node(), 0)
    while not fringe.empty():
        fringe_var = fringe.get()
        node, depth = fringe_var.node, fringe_var.depth
        if node.program in visited or depth > max_depth:
            continue
        visited.add(node.program)
        if g.is_goal_node(node):
            yield node
        for child in g.expand_node(node):
            add_to_fringe(child, depth + 1)


@dataclass(order=True)
class BoundedAStarNode:
    """
    Represents a node in the A* search tree.
    """

    cost: float
    node: SExpression = field(compare=False)
    depth: int = field(compare=True)



"""
DSL of the form:

# one :: float
# ones :: tensor[float]
# int_int_add :: float -> float -> float
# Tint_int_add :: tensor[float] -> float -> tensor[float]
# Tint_Tint_add :: tensor[float] -> tensor[float] -> tensor[float]
# Function signature for Tint_linear
# Tint_linear :: LinearObj -> tensor[float] -> tensor[float]
# Linear_c :: LinearQbj
"""
import torch
from torch import nn

from neurosym.dsl.dsl_factory import DSLFactory


def differentiable_arith_dsl(length):
    dslf = DSLFactory(L=length)
    dslf.concrete("one", "() -> f", lambda: torch.tensor(1.0))
    dslf.concrete("ones", "() -> {f, $L}", lambda: torch.ones(length))
    dslf.concrete("int_int_add", "(f, f) -> f", lambda x, y: x + y)
    dslf.concrete("Tint_int_add", "({f, $L}, f) -> {f, $L}", lambda x, y: x + y)
    dslf.concrete("Tint_Tint_add", "({f, $L}, {f, $L}) -> {f, $L}", lambda x, y: x + y)
    dslf.concrete(
        "app_Tint", "({f, $L} -> {f, $L}, {f, $L}) -> {f, $L}", lambda f, x: f(x)
    )
    dslf.parameterized(
        "Linear_c",
        "() -> {f, $L} -> {f, $L}",
        lambda linear: linear,
        dict(linear=lambda: nn.Linear(length, length)),
    )
    return dslf.finalize()



import unittest

import neurosym as ns

dslf = ns.DSLFactory()
dslf.concrete("1", "() -> i -> i", lambda: 1)
dslf.concrete("+", "(#t -> i, #t -> i) -> #t -> i", lambda x, y: lambda t: x(t) + y(t))
dslf.concrete("id", "#a -> #a", lambda x: x)
dslf.concrete(
    "compose", "(#a -> #b, #b -> #c) -> #a -> #c", lambda f, g: lambda x: f(g(x))
)
dsl = dslf.finalize()


class TestEnumeratability(unittest.TestCase):
    def test_dsl_productions(self):
        expected = """
                1 :: () -> i -> i
                + :: (#t -> i, #t -> i) -> #t -> i
                id :: #a -> #a
        compose_0 :: (#a -> () -> () -> i, (() -> () -> i) -> #c) -> #a -> #c
        compose_1 :: (#a -> () -> i, (() -> i) -> #c) -> #a -> #c
        compose_2 :: (#a -> (i, i) -> i, ((i, i) -> i) -> #c) -> #a -> #c
        compose_3 :: (#a -> i -> i, (i -> i) -> #c) -> #a -> #c
        compose_4 :: (#a -> i, i -> #c) -> #a -> #c
        """
        actual = dsl.render()
        self.assertEqual(
            {line.strip() for line in actual.strip().split("\n")},
            {line.strip() for line in expected.strip().split("\n")},
        )

    def test_basic_enumerate(self):
        expans = {
            ns.render_s_expression(prog)
            for prog in dsl.expansions_for_type(
                ns.TypeWithEnvironment(ns.parse_type("i -> i"), ns.Environment.empty())
            )
        }
        print(expans)
        self.assertSetEqual(
            expans,
            {
                "(1)",
                "(id ??::<i -> i>)",
                "(+ ??::<i -> i> ??::<i -> i>)",
                "(compose_0 ??::<i -> () -> () -> i> ??::<(() -> () -> i) -> i>)",
                "(compose_1 ??::<i -> () -> i> ??::<(() -> i) -> i>)",
                "(compose_2 ??::<i -> (i, i) -> i> ??::<((i, i) -> i) -> i>)",
                "(compose_3 ??::<i -> i -> i> ??::<(i -> i) -> i>)",
                "(compose_4 ??::<i -> i> ??::<i -> i>)",
            },
        )

    def test_filtered_variable(self):
        dslf_2 = ns.DSLFactory()
        dslf_2.concrete("1", "() -> i", lambda: 1)
        dslf_2.concrete("1f", "() -> f", lambda: 1)
        dslf_2.filtered_type_variable(
            "num", lambda x: isinstance(x, ns.AtomicType) and x.name in ["i", "f"]
        )
        dslf_2.concrete("+", "(%num, %num) -> %num", lambda x, y: x + y)
        dslf_2.concrete("*", "(#a, #a) -> #a", lambda x, y: x + y)
        dsl_2 = dslf_2.finalize()
        expans = {
            ns.render_s_expression(prog)
            for prog in dsl_2.expansions_for_type(
                ns.TypeWithEnvironment(ns.parse_type("i"), ns.Environment.empty())
            )
        }
        print(expans)
        self.assertSetEqual(
            expans, {"(* ??::<i> ??::<i>)", "(+ ??::<i> ??::<i>)", "(1)"}
        )
        expans = {
            ns.render_s_expression(prog)
            for prog in dsl_2.expansions_for_type(
                ns.TypeWithEnvironment(ns.parse_type("() -> i"), ns.Environment.empty())
            )
        }
        print(expans)
        self.assertSetEqual(expans, {"(* ??::<() -> i> ??::<() -> i>)"})



from dataclasses import dataclass
from types import NoneType
from typing import Callable, Dict, List, Tuple, Union

from neurosym.types.type_with_environment import (
    Environment,
    PermissiveEnvironmment,
    TypeWithEnvironment,
)
from neurosym.utils.tree_trie import TreeTrie

from ..programs.hole import Hole
from ..programs.s_expression import InitializedSExpression, SExpression
from ..types.type import GenericTypeVariable, Type
from .production import Production


@dataclass
class DSL:
    # list of productions
    productions: List[Production]
    # list of types that can be used as the root of a program. If None, all types can be
    # used as the root.
    valid_root_types: Union[List[Type], NoneType]
    # max depth of a valid type for a program in this DSL
    max_type_depth: float
    # max depth of a valid environment for a program in this DSL
    max_env_depth: float

    def __post_init__(self):
        symbols = set()
        for production in self.productions:
            assert (
                production.symbol() not in symbols
            ), f"Duplicate symbol {production.symbol()}"
            symbols.add(production.symbol())
        self._production_by_symbol = {
            production.symbol(): production for production in self.productions
        }

        self._out_type_to_prod_idx = TreeTrie.empty()
        for i, prod in enumerate(self.productions):
            self._out_type_to_prod_idx.insert(
                prod.type_signature().return_type_template(),
                i,
                is_wildcard_predicate=lambda x: isinstance(x, GenericTypeVariable),
            )

    def symbols(self):
        return self._production_by_symbol.keys()

    def arity(self, sym: str) -> int:
        """
        Returns the arity of the production with the given symbol.
        """
        return self.get_production(sym).type_signature().arity()

    def productions_for_type(
        self, typ: TypeWithEnvironment
    ) -> List[Tuple[Production, List[TypeWithEnvironment]]]:
        for idx in sorted(self._out_type_to_prod_idx.query(typ.typ)):
            production = self.productions[idx]
            arg_types = production.type_signature().unify_return(typ)
            if arg_types is not None:
                yield production, arg_types

    def expansions_for_type(self, typ: TypeWithEnvironment) -> List[SExpression]:
        """
        Possible expansions for the given type.

        An expansion is an SExpression with holes in it. The holes can be filled in with
        other SExpressions to produce a complete SExpression.
        """
        return [
            SExpression(
                production.symbol(),
                tuple(Hole.of(t) for t in arg_types),
            )
            for production, arg_types in self.productions_for_type(typ)
        ]

    def get_production(self, symbol: str) -> Production:
        """
        Return the production with the given symbol.
        """
        assert isinstance(symbol, str), f"Expected string, got {type(symbol)}: {symbol}"
        return self._production_by_symbol[symbol]

    def initialize(self, program: SExpression) -> InitializedSExpression:
        """
        Initializes all the productions in the given program.

        Returns a new program with the same structure, but with all the productions
        initialized.
        """
        if hasattr(program, "__initialize__"):
            return program.__initialize__(self)
        prod = self.get_production(program.symbol)
        return InitializedSExpression(
            program.symbol,
            tuple(self.initialize(child) for child in program.children),
            prod.initialize(self),
        )

    def compute(self, program: InitializedSExpression):
        if hasattr(program, "__compute_value__"):
            return program.__compute_value__(self)
        prod = self.get_production(program.symbol)
        return prod.apply(self, program.state, program.children)

    def all_rules(
        self, care_about_variables, valid_root_types=None
    ) -> Dict[Type, List[Tuple[str, List[Type]]]]:
        """
        Returns a dictionary of all the rules in the DSL, where the keys are the types
        that can be expanded, and the values are a list of tuples of the form
        (symbol, types), where symbol is the symbol of the production, and types is a
        list of types that can be used to expand the given type.

        This is useful for generating a PCFG.
        """
        if valid_root_types is None:
            if self.valid_root_types is not None:
                valid_root_types = self.valid_root_types
            else:
                raise ValueError(
                    "Cannot generate rules for a DSL with no valid root types."
                )
        twes_to_expand = [
            TypeWithEnvironment(
                type,
                (
                    Environment.empty()
                    if care_about_variables
                    else PermissiveEnvironmment()
                ),
            )
            for type in valid_root_types
        ]
        rules = {}
        while len(twes_to_expand) > 0:
            twe = twes_to_expand.pop()
            if (
                twe.typ.depth > self.max_type_depth
                or len(twe.env) > self.max_env_depth
                or twe in rules
            ):
                continue
            rules[twe] = []
            for prod, twes in self.productions_for_type(twe):
                rules[twe].append((prod.symbol(), twes))
                twes_to_expand.extend(twes)
        if not care_about_variables:
            rules = {
                out_twe.typ: [
                    (sym, [inp_twe.typ for inp_twe in inp_twes])
                    for sym, inp_twes in rules
                ]
                for out_twe, rules in rules.items()
            }
        return rules

    def constructible_symbols(self, care_about_variables, valid_root_types=None):
        """
        Returns all the symbols that can be constructed from the given target types.
        """
        type_to_rules = self.all_rules(
            care_about_variables=care_about_variables,
            valid_root_types=valid_root_types,
        )

        constructible = set()

        while True:
            done = True
            for out_t, rules in type_to_rules.items():
                if out_t in constructible:
                    continue
                for _, in_t in rules:
                    if set(in_t).issubset(constructible):
                        constructible.add(out_t)
                        done = False
            if done:
                break
        return {
            sym
            for _, rules in type_to_rules.items()
            for sym, in_t in rules
            if all(t in constructible for t in in_t)
        }

    def compute_type(
        self,
        program: SExpression,
        lookup: Callable[[SExpression], TypeWithEnvironment] = lambda x: None,
    ) -> TypeWithEnvironment:
        """
        Computes the type of the given program.
        """
        if lookup is not None:
            res = lookup(program)
            if res is not None:
                return res

        child_types = [self.compute_type(child, lookup) for child in program.children]
        prod = self.get_production(program.symbol)
        return prod.type_signature().unify_arguments(child_types)

    def render(self) -> str:
        """
        Render this DSL as a string.
        """
        return "\n".join(production.render() for production in self.productions)

    def add_production(self, prod):
        return DSL(
            self.productions + [prod],
            self.valid_root_types,
            self.max_type_depth,
            self.max_env_depth,
        )

    def with_valid_root_types(self, valid_root_types):
        return DSL(
            self.productions,
            valid_root_types,
            self.max_type_depth,
            self.max_env_depth,
        )



from abc import abstractmethod
from dataclasses import dataclass
from functools import cached_property
from types import NoneType
from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from neurosym.program_dist.distribution import (
    ProgramDistribution,
    ProgramDistributionFamily,
)
from neurosym.program_dist.enumeration_chunk_size import DEFAULT_CHUNK_SIZE
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.programs.s_expression import SExpression


@dataclass
class TreeDistribution:
    """
    Distribution over SExpressions as trees.

    Internally, we represent the productions in the language as integers, which we
        call indices.
    """

    limit: int
    # input: tuple of tuples of (ancestor index, position)
    #        which is the path to the current node, with the
    #        most immediate ancestor at the end.
    # output: list of (production index, likelihood) pairs
    distribution: Dict[Tuple[Tuple[int, int], ...], List[Tuple[int, float]]]
    # production index -> (symbol, arity). at 0 should be the root.
    symbols: List[Tuple[str, int]]
    # Preorder mask constructor
    mask_constructor: Callable[["TreeDistribution"], PreorderMask]

    @cached_property
    def symbol_to_index(self) -> Dict[str, int]:
        return {symbol: i for i, (symbol, _) in enumerate(self.symbols)}

    @cached_property
    def distribution_dict(self) -> Dict[Tuple[Tuple[int, int], ...], Dict[int, float]]:
        return {k: dict(v) for k, v in self.distribution.items()}

    @cached_property
    def likelihood_arrays(
        self,
    ) -> Dict[Tuple[Tuple[int, int], ...], Tuple[np.ndarray, np.ndarray]]:
        return {
            k: (
                np.array([x[0] for x in v]),
                np.array([x[1] for x in v]),
            )
            for k, v in self.distribution.items()
        }

    @cached_property
    def sampling_dict_arrays(
        self,
    ) -> Dict[Tuple[Tuple[int, int], ...], Tuple[np.ndarray, np.ndarray]]:
        return {
            k: (syms, np.exp(log_probs))
            for k, (syms, log_probs) in self.likelihood_arrays.items()
        }


class TreeProgramDistributionFamily(ProgramDistributionFamily):
    """
    See `tree_dist_enumerator.py` for more information.
    """

    @abstractmethod
    def compute_tree_distribution(
        self, distribution: Union[ProgramDistribution, NoneType]
    ) -> TreeDistribution:
        """
        Returns a tree distribution representing the given program distribution.

        If `distribution` is `None`, returns a tree distribution with all fields
            initialized except `distribution`. This is useful for tasks where you want
            the skeleton of the tree distribution, but don't need the actual distribution.
        """

    def tree_distribution(self, distribution: ProgramDistribution) -> TreeDistribution:
        """
        Cached version of `compute_tree_distribution`.
        """
        # This is a bit of a hack, but it reduces the need to pass around
        # the tree distribution everywhere, or to compute it multiple times.
        # pylint: disable=protected-access
        if not hasattr(distribution, "_tree_distribution"):
            distribution._tree_distribution = self.compute_tree_distribution(
                distribution
            )
        return distribution._tree_distribution

    @cached_property
    def tree_distribution_skeleton(self) -> TreeDistribution:
        """
        Cached version of `compute_tree_distribution(None)`.
        """

        return self.compute_tree_distribution(None)

    def enumerate(
        self,
        dist: ProgramDistribution,
        *,
        min_likelihood: float = float("-inf"),
        chunk_size: float = DEFAULT_CHUNK_SIZE,
    ):
        # pylint: disable=cyclic-import
        from neurosym.program_dist.tree_distribution.tree_dist_enumerator import (
            enumerate_tree_dist,
        )

        tree_dist = self.tree_distribution(dist)

        return enumerate_tree_dist(
            tree_dist, min_likelihood=min_likelihood, chunk_size=chunk_size
        )

    def compute_likelihood(
        self, dist: ProgramDistribution, program: SExpression
    ) -> float:
        """
        Compute the likelihood of a program under a distribution.
        """
        from .tree_dist_likelihood_computer import compute_likelihood

        dist = self.tree_distribution(dist)
        preorder_mask = dist.mask_constructor(dist)
        preorder_mask.on_entry(0, 0)
        return compute_likelihood(dist, program, ((0, 0),), preorder_mask)

    def sample(
        self,
        dist: ProgramDistribution,
        rng: np.random.RandomState,
        *,
        depth_limit=float("inf"),
    ) -> SExpression:
        # pylint: disable=cyclic-import

        from neurosym.program_dist.tree_distribution.tree_dist_sampler import (
            sample_tree_dist,
        )

        tree_dist = self.tree_distribution(dist)
        element = sample_tree_dist(tree_dist, rng, depth_limit=depth_limit)
        assert element.symbol == "<root>"
        [element] = element.children
        return element



from dataclasses import dataclass
from typing import List

from ..types.type_with_environment import TypeWithEnvironment
from .s_expression import SExpression


@dataclass(eq=True, frozen=True)
class Hole:
    @classmethod
    def of(cls, twe: TypeWithEnvironment) -> "Hole":
        assert isinstance(twe, TypeWithEnvironment)
        cls._id = getattr(cls, "_id", 0) + 1
        return cls(id=cls._id, twe=twe)

    id: int
    twe: TypeWithEnvironment

    def __to_pair__(self, for_stitch: bool) -> str:
        from neurosym.types.type_string_repr import render_type

        del for_stitch
        env_short = self.twe.env.short_repr()
        if env_short:
            env_short = "|" + env_short
        return f"??::<{render_type(self.twe.typ)}{env_short}>"


def replace_holes(
    program: SExpression, holes: List[Hole], hole_replacements: List[SExpression]
) -> SExpression:
    """
    Replace the holes in node with the hole_replacements in the given SExpression.

    :param program: the SExpression to replace holes in
    :param holes: the holes to replace
    :param hole_replacements: the replacements for the holes
    """

    assert len(holes) == len(hole_replacements)
    if isinstance(program, Hole):
        if program not in holes:
            return program
        return hole_replacements[holes.index(program)]
    return SExpression(
        program.symbol,
        tuple(
            replace_holes(child, holes, hole_replacements) for child in program.children
        ),
    )


def all_holes(program: SExpression) -> List[Hole]:
    """
    Yield all holes in the given SExpression.

    :param program: the SExpression to find holes in
    """
    if isinstance(program, Hole):
        yield program
    else:
        for child in program.children:
            yield from all_holes(child)



import numpy as np


def compute_likelihood(tree_dist, program, parents, preorder_mask):
    syms, log_probs = tree_dist.likelihood_arrays[parents]
    start_position = parents[-1][1]
    mask = preorder_mask.compute_mask(start_position, syms)
    denominator = np.logaddexp.reduce(log_probs[mask])
    top_symbol = tree_dist.symbol_to_index[program.symbol]
    likelihood = (
        tree_dist.distribution_dict[parents].get(top_symbol, -float("inf"))
        - denominator
    )
    if likelihood == -float("inf"):
        return -float("inf")
    preorder_mask.on_entry(start_position, top_symbol)
    for i, child in enumerate(program.children):
        likelihood += compute_likelihood(
            tree_dist,
            child,
            (parents + ((top_symbol, i),))[-tree_dist.limit :],
            preorder_mask=preorder_mask,
        )
    preorder_mask.on_exit(start_position, top_symbol)
    return likelihood



import unittest
from functools import lru_cache

import numpy as np

import neurosym as ns

dsl = ns.examples.mutable_arith_combinators_dsl

out_t = ns.parse_type("(i) -> i")


@lru_cache(maxsize=None)
def corpus():
    fam = ns.BigramProgramDistributionFamily(dsl, valid_root_types=[out_t])
    dist = fam.uniform()
    return sorted(
        {
            fam.sample(dist, np.random.RandomState(i), depth_limit=20)
            for i in range(100)
        },
        key=str,
    )


class BasicDSLTest(unittest.TestCase):
    def test_independent_mutability(self):
        prog = ns.parse_s_expression("(ite (even? (x)) (count) (count))")
        fn_1 = dsl.compute(dsl.initialize(prog))
        fn_2 = dsl.compute(dsl.initialize(prog))
        self.assertEqual([fn_1(2), fn_1(4), fn_1(8)], [1, 2, 3])
        self.assertEqual([fn_1(2), fn_1(4), fn_1(8)], [4, 5, 6])
        # fn_2 is independent of fn_1
        self.assertEqual([fn_2(2), fn_2(4), fn_2(8)], [1, 2, 3])
        self.assertEqual([fn_1(1), fn_1(2), fn_1(3)], [1, 7, 2])
        self.assertEqual([fn_1(1), fn_1(2), fn_1(3)], [3, 8, 4])


class CompressionTest(unittest.TestCase):
    def fuzzy_check_fn_same(self, fn_1, fn_2):
        inputs = np.random.RandomState(0).randint(100, size=100)
        outputs_1 = [fn_1(x) for x in inputs]
        outputs_2 = [fn_2(x) for x in inputs]
        self.assertEqual(outputs_1, outputs_2)

    def test_single_step(self):
        dsl2, rewritten = ns.compression.single_step_compression(dsl, corpus())
        self.assertEqual(len(rewritten), len(corpus()))
        for orig, rewr in zip(corpus(), rewritten):
            self.fuzzy_check_fn_same(
                dsl.compute(dsl.initialize(orig)),
                dsl2.compute(dsl2.initialize(rewr)),
            )

    def test_multi_step(self):
        dsl2, rewritten = ns.compression.multi_step_compression(dsl, corpus(), 5)
        self.assertEqual(len(rewritten), len(corpus()))
        for orig, rewr in zip(corpus(), rewritten):
            self.fuzzy_check_fn_same(
                dsl.compute(dsl.initialize(orig)),
                dsl2.compute(dsl2.initialize(rewr)),
            )


class BasicProcessDSL(unittest.TestCase):
    def setUp(self):
        self.dsl = ns.examples.basic_arith_dsl(True)
        print(ns.examples.basic_arith_dsl(True).render())
        self.fn_code = "(lam_0 (+ ($1_0) ($0_0)))"

    def test_compute(self):
        fn = self.dsl.compute(self.dsl.initialize(ns.parse_s_expression(self.fn_code)))
        self.assertEqual(fn(1, 2), 3)
        self.assertEqual(fn(3, 4), 7)
        self.assertEqual(fn(1000, -24), 976)

    def test_basic_compress(self):
        code = [
            "(lam_0 (+ (+ (1) (1)) ($0_0)))",
            "(lam_0 (+ (1) ($0_0)))",
        ]
        code = [ns.parse_s_expression(x) for x in code]
        dsl2, rewritten = ns.compression.single_step_compression(self.dsl, code)
        self.assertEqual(len(rewritten), len(code))
        self.assertEqual(
            [ns.render_s_expression(x) for x in rewritten],
            [
                "(__10 (+ (1) (1)))",
                "(__10 (1))",
            ],
        )
        self.assertEqual(
            dsl2.productions[-1].render().strip(),
            "__10 :: i -> (i, i) -> i = (lam-abstr (#0) (lam_0 (+ #0 ($0_0))))",
        )



from dataclasses import dataclass
from typing import Dict, Tuple


@dataclass(frozen=True, eq=True)
class SExpression:
    symbol: str
    children: Tuple["SExpression"]

    @property
    def postorder(self):
        for x in self.children:
            yield from x.postorder
        yield self


@dataclass
class InitializedSExpression:
    symbol: str
    children: Tuple["InitializedSExpression"]
    # state includes things related to the execution of the program,
    # e.g. weights of a neural network
    state: Dict[str, object]

    def all_state_values(self):
        yield from self.state.values()
        for child in self.children:
            yield from child.all_state_values()



from types import NoneType
from typing import Union

from neurosym.types.type import (
    ArrowType,
    AtomicType,
    FilteredTypeVariable,
    ListType,
    TensorType,
    TypeVariable,
)
from neurosym.types.type_signature import FunctionTypeSignature

SPECIAL_CHARS = ["{", "}", "[", "]", "(", ")", "->", ","]


class TypeDefiner:
    def __init__(self, **env):
        self.env = env
        self.filters = {}

    def __call__(self, type_str):
        return parse_type(type_str, self)

    def sig(self, type_str):
        typ = self(type_str)
        return FunctionTypeSignature(list(typ.input_type), typ.output_type)

    def typedef(self, key, type_str):
        self.env[key] = self(type_str)

    def filtered_type_variable(self, key, type_filter):
        self.filters[key] = type_filter

    def lookup_type(self, key):
        return self.env[key]

    def lookup_filter(self, key):
        return self.filters[key]


def render_type(t):
    if isinstance(t, AtomicType):
        return t.name
    if isinstance(t, TypeVariable):
        return "#" + t.name
    if isinstance(t, FilteredTypeVariable):
        return "%" + t.name
    if isinstance(t, TensorType):
        return "{" + ", ".join([render_type(t.dtype), *map(str, t.shape)]) + "}"
    if isinstance(t, ListType):
        return "[" + render_type(t.element_type) + "]"
    if isinstance(t, ArrowType):
        if len(t.input_type) == 1 and not isinstance(t.input_type[0], ArrowType):
            return render_type(t.input_type[0]) + " -> " + render_type(t.output_type)
        return (
            "("
            + ", ".join(map(render_type, t.input_type))
            + ") -> "
            + render_type(t.output_type)
        )
    raise NotImplementedError(f"Unknown type {t}")


def parse_type_from_buf(reversed_buf, env: TypeDefiner):
    assert isinstance(env, TypeDefiner)
    first_tok = reversed_buf.pop()
    if first_tok.isnumeric():
        return int(first_tok)
    if first_tok.startswith("$"):
        return env.lookup_type(first_tok[1:])
    if first_tok == "{":
        internal_type = parse_type_from_buf(reversed_buf, env)
        shape = []
        while True:
            tok = reversed_buf.pop()
            if tok == "}":
                break
            assert tok == ",", f"Expected ',' but got {tok}"
            size = parse_type_from_buf(reversed_buf, env)
            shape.append(size)
        return TensorType(internal_type, tuple(shape))
    if first_tok == "[":
        internal_type = parse_type_from_buf_multi(reversed_buf, env)
        close_bracket = reversed_buf.pop()
        assert close_bracket == "]", f"Expected ']' but got {close_bracket}"
        return ListType(internal_type)
    if first_tok == "(":
        input_types = []
        while True:
            if reversed_buf[-1] == ")":
                reversed_buf.pop()
                break
            input_types.append(parse_type_from_buf_multi(reversed_buf, env))
            tok = reversed_buf.pop()
            if tok == ")":
                break
            assert tok == ",", f"Expected ',' but got {tok}"
        tok = reversed_buf.pop()
        assert tok == "->", f"Expected '->' but got {tok}"
        output_type = parse_type_from_buf_multi(reversed_buf, env)
        return ArrowType(tuple(input_types), output_type)
    if first_tok.startswith("#"):
        return TypeVariable(first_tok[1:])
    if first_tok.startswith("%"):
        name = first_tok[1:]
        return FilteredTypeVariable(name, type_filter=env.lookup_filter(name))
    return AtomicType(first_tok)


def parse_type_from_buf_multi(reversed_buf, env):
    t_head = parse_type_from_buf(reversed_buf, env)
    if not reversed_buf:
        return t_head
    if reversed_buf and reversed_buf[-1] != "->":
        return t_head
    reversed_buf.pop()
    t_tail = parse_type_from_buf_multi(reversed_buf, env)
    return ArrowType((t_head,), t_tail)


def lex(s):
    buf = []
    for c in s:
        if c in SPECIAL_CHARS:
            buf.append(c)
        elif c == " ":
            buf.append("")
        else:
            if len(buf) > 0 and buf[-1] not in SPECIAL_CHARS:
                buf[-1] += c
            else:
                buf.append(c)
    return [tok for tok in buf if tok != ""]


def parse_type(s, env: Union[TypeDefiner, NoneType] = None):
    if env is None:
        env = TypeDefiner()
    assert isinstance(env, TypeDefiner)
    buf = lex(s)[::-1]
    t = parse_type_from_buf_multi(buf, env)
    assert buf == [], f"Extra tokens {buf[::-1]}"
    return t



from typing import Callable, Iterator, List, Tuple

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def collect_preorder_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    mask: Callable[[TreeDistribution], PreorderMask],
) -> Iterator[Tuple[SExpression, List[str]]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution.
    """
    mask = mask(tree_dist)
    mask.on_entry(0, 0)
    yield from collect_preorder_symbols_dfs(s_exp, tree_dist, mask, 0)


def collect_preorder_symbols_dfs(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    mask: PreorderMask,
    position: int,
) -> Iterator[Tuple[SExpression, List[str]]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution.
    """
    idxs = np.arange(len(tree_dist.symbols))
    bool_mask = mask.compute_mask(position, idxs)
    alts = tuple(int(x) for x in idxs[bool_mask])
    yield s_exp, alts
    mask.on_entry(position, tree_dist.symbol_to_index[s_exp.symbol])
    for idx, child in enumerate(s_exp.children):
        yield from collect_preorder_symbols_dfs(child, tree_dist, mask, idx)
    mask.on_exit(position, tree_dist.symbol_to_index[s_exp.symbol])


def annotate_with_alternate_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    mask: Callable[[TreeDistribution], PreorderMask],
    summary_fn=lambda chosen, alts: f"{chosen}/{','.join(sorted(alts))}",
) -> SExpression:
    """
    Annotates the S-Expression with the alternate symbols that could have been
    selected in the tree distribution.
    """
    node_id_to_alts = {
        id(node): tuple(tree_dist.symbols[alt][0] for alt in alts)
        for node, alts in collect_preorder_symbols(s_exp, tree_dist, mask)
    }

    def replace(s):
        symbol = summary_fn(s.symbol, node_id_to_alts[id(s)])
        children = [replace(c) for c in s.children]
        return SExpression(symbol, children)

    return replace(s_exp)



"""
Just checks that the package can be imported
"""

import unittest

import neurosym as ns

dsl = ns.examples.basic_arith_dsl()


class TestSearch(unittest.TestCase):
    def test_bfs(self):
        g = ns.DSLSearchGraph(
            dsl,
            ns.parse_type("i"),
            ns.ChooseFirst(),
            lambda x: dsl.compute(dsl.initialize(x.program)) == 4,
            metadata_computer=ns.NoMetadataComputer(),
        )
        node = next(ns.search.bfs(g)).program
        self.assertEqual(
            node,
            ns.SExpression(
                symbol="+",
                children=(
                    ns.SExpression(
                        symbol="+",
                        children=(
                            ns.SExpression(
                                symbol="+",
                                children=(
                                    ns.SExpression(symbol="1", children=()),
                                    ns.SExpression(symbol="1", children=()),
                                ),
                            ),
                            ns.SExpression(symbol="1", children=()),
                        ),
                    ),
                    ns.SExpression(symbol="1", children=()),
                ),
            ),
        )

    def test_astar(self):
        g = ns.DSLSearchGraph(
            dsl,
            ns.parse_type("i"),
            ns.ChooseFirst(),
            lambda x: dsl.compute(dsl.initialize(x.program)) == 4,
            metadata_computer=ns.NoMetadataComputer(),
        )

        def cost(x):
            if isinstance(x.program, ns.SExpression) and x.program.children:
                return len(str(x.program.children[0]))
            return 0

        node = next(ns.search.astar(g, cost)).program
        print(node)
        self.assertEqual(
            node,
            ns.SExpression(
                symbol="+",
                children=(
                    ns.SExpression(symbol="1", children=()),
                    ns.SExpression(
                        symbol="+",
                        children=(
                            ns.SExpression(
                                symbol="+",
                                children=(
                                    ns.SExpression(symbol="1", children=()),
                                    ns.SExpression(symbol="1", children=()),
                                ),
                            ),
                            ns.SExpression(symbol="1", children=()),
                        ),
                    ),
                ),
            ),
        )



from torch import nn


class TorchProgramModule(nn.Module):
    def __init__(self, dsl, program):
        super().__init__()
        self.dsl = dsl
        self.initalized_program = dsl.initialize(program)
        self.contained_modules = nn.ModuleList(
            list(self.initalized_program.all_state_values())
        )

    def forward(self, *args):
        return self.dsl.compute(self.initalized_program)(*args)



from neurosym.search_graph.search_graph import SearchGraph


def bfs(g: SearchGraph, iteration_limit=float("inf")):
    """
    Performs a breadth-first search on the given search graph, yielding each node in the
    order it was visited.

    :param g: Search graph to search over
    """
    visited = set()
    queue = [g.initial_node()]
    while queue:
        node = queue.pop(0)
        iteration_limit -= 1
        if iteration_limit < 0:
            break
        if node in visited:
            continue
        visited.add(node)
        if g.is_goal_node(node):
            yield node
        for child in g.expand_node(node):
            queue.append(child)



import unittest
from fractions import Fraction

import numpy as np
import torch

import neurosym as ns
from tests.utils import assertDSL

from .utils import ProbabilityTester


def get_dsl(with_vars=False):
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    if with_vars:
        dslf.concrete("call", "(i -> i, i) -> i", lambda f, x: f(x))
        dslf.lambdas()
    dslf.prune_to("i")
    return dslf.finalize()


dsl = get_dsl()
fam = ns.BigramProgramDistributionFamily(dsl)
dsl_with_vars = get_dsl(with_vars=True)
fam_with_vars = ns.BigramProgramDistributionFamily(dsl_with_vars)

uniform = [
    [
        # root -> each
        [0, 1 / 3, 1 / 3, 1 / 3],
        [0] * 4,
    ],
    [
        # + -> each first arg
        [0, 1 / 3, 1 / 3, 1 / 3],
        # + -> each second arg
        [0, 1 / 3, 1 / 3, 1 / 3],
    ],
    [[0] * 4] * 2,
    [[0] * 4] * 2,
]


class BigramSamplerTest(ProbabilityTester):
    def test_uniform_sample_counts(self):
        n = 10_000
        pdelta = 0.015
        dist = fam.uniform()
        self.assertSameProbDist(dist, uniform)
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 3, pdelta, samples.count("(1)"))
        for plus_two_things in (
            "(+ (1) (1))",
            "(+ (1) (2))",
            "(+ (2) (1))",
            "(+ (2) (2))",
        ):
            self.assertBinomial(
                n, 1 / 3 * 1 / 3 * 1 / 3, pdelta, samples.count(plus_two_things)
            )


class BigramParameterShapeTest(unittest.TestCase):
    def test_shape(self):
        self.assertEqual(fam.parameters_shape(), (4, 2, 4))


class BigramWithParametersGetParametersTest(ProbabilityTester):
    def test_produces_uniform_with_fixed_input(self):
        parameters = torch.zeros((4, 2, 4))
        for off in -1, 298, 29, 3:
            self.assertSameProbDist(
                fam.with_parameters((parameters + off)[None])[0], uniform
            )

    def test_irrelevant_entries_dont_matter(self):
        parameters = torch.zeros((4, 2, 4))
        parameters[0, 0, 0] = 1038
        print(parameters)
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[0, 1, 1] = -3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[2] = 3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

    def test_can_make_deterministic(self):
        parameters = torch.zeros((4, 2, 4))
        # + -> always 1 then 2
        parameters[1, 0, 2] = 100
        parameters[1, 1, 3] = 100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)", "(+ (1) (2))"})

        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(+ (1) (2))"))

    def test_can_make_impossible(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(2)"))

    def test_can_weight_precisely(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100
        # make root -> 1 twice as likely as root -> 2
        parameters[0, 0, 2] = np.log(2)

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 2 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))

    def test_fam_with_variables_dsl(self):
        assertDSL(
            self,
            dsl_with_vars.render(),
            """
            $0_0 :: V<i@0>
            $1_0 :: V<i@1>
            $2_0 :: V<i@2>
            $3_0 :: V<i@3>
            + :: (i, i) -> i
            1 :: () -> i
            2 :: () -> i
            call :: (i -> i, i) -> i
            lam :: L<#body|i> -> i -> #body
            """,
        )

    def test_sample_with_variables(self):
        dist = fam_with_vars.uniform()
        n = 20000
        samples = [
            fam_with_vars.sample(dist, np.random.RandomState(i)) for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(1)"))
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(2)"))
        # see test_call_with_variables for the math
        self.assertBinomial(n, 1 / 80, 0.01, samples.count("(call (lam ($0_0)) (1))"))


class BigramCountProgramsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_counts_single_program(self):
        data = [[ns.parse_s_expression("(+ (1) (2))")]]
        counts = fam.count_programs(data)
        self.assertEqual(
            counts,
            ns.BigramProgramCountsBatch(
                [
                    ns.BigramProgramCounts(
                        {
                            # root -> +
                            ((0, 0),): {1: 1},
                            # + -> 1 as the first arg
                            ((1, 0),): {2: 1},
                            # + -> 2 as the second arg
                            ((1, 1),): {3: 1},
                        },
                        {
                            # root -> ?
                            ((0, 0),): {(1, 2, 3): 1},
                            # + -> ?
                            ((1, 0),): {(1, 2, 3): 1},
                            ((1, 1),): {(1, 2, 3): 1},
                        },
                    )
                ]
            ),
        )

    def test_counts_multiple_programs(self):
        data = [
            [ns.parse_s_expression(x) for x in ("(+ (1) (2))", "(+ (1) (1))")],
        ]
        counts = fam.count_programs(data)
        self.assertEqual(
            counts,
            ns.BigramProgramCountsBatch(
                [
                    ns.BigramProgramCounts(
                        {
                            # root -> +
                            ((0, 0),): {1: 2},
                            # + -> 1 as the first arg (twice)
                            ((1, 0),): {2: 2},
                            # + -> 1 as the second arg; + -> 2 as the second arg
                            ((1, 1),): {2: 1, 3: 1},
                        },
                        {
                            # root -> ?
                            ((0, 0),): {(1, 2, 3): 2},
                            # + -> ?
                            ((1, 0),): {(1, 2, 3): 2},
                            ((1, 1),): {(1, 2, 3): 2},
                        },
                    )
                ]
            ),
        )

    def test_counts_separate_programs(self):
        data = [
            [ns.parse_s_expression("(+ (1) (2))")],
            [ns.parse_s_expression("(+ (1) (1))")],
        ]
        counts = fam.count_programs(data)
        np.testing.assert_equal(
            counts,
            ns.BigramProgramCountsBatch(
                [
                    ns.BigramProgramCounts(
                        {
                            # root -> +
                            ((0, 0),): {1: 1},
                            # + -> 1 as the first arg
                            ((1, 0),): {2: 1},
                            # + -> 2 as the second arg
                            ((1, 1),): {3: 1},
                        },
                        {
                            # root -> ?
                            ((0, 0),): {(1, 2, 3): 1},
                            # + -> ?
                            ((1, 0),): {(1, 2, 3): 1},
                            ((1, 1),): {(1, 2, 3): 1},
                        },
                    ),
                    ns.BigramProgramCounts(
                        {
                            # root -> +
                            ((0, 0),): {1: 1},
                            # + -> 1 as the first arg
                            ((1, 0),): {2: 1},
                            # + -> 1 as the second arg
                            ((1, 1),): {2: 1},
                        },
                        {
                            # root -> ?
                            ((0, 0),): {(1, 2, 3): 1},
                            # + -> ?
                            ((1, 0),): {(1, 2, 3): 1},
                            ((1, 1),): {(1, 2, 3): 1},
                        },
                    ),
                ]
            ),
        )


class BigramParameterDifferenceLossTest(unittest.TestCase):
    def computeLoss(self, logits, programs, family=fam):
        programs = [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        print(programs)
        loss = family.parameter_difference_loss(
            logits,
            family.count_programs(programs),
        )
        loss = loss.detach().cpu().numpy()
        return loss

    def assertLoss(self, logits, programs, target, **kwargs):
        np.testing.assert_almost_equal(
            self.computeLoss(logits, programs, **kwargs), target
        )

    def assertUniformLoss(self, programs, target, **kwargs):
        logits = torch.zeros((1, 4, 2, 4))
        self.assertLoss(logits, programs, target, **kwargs)
        # fill in impossible location +, 0 -> <root>
        logits[0, 1, 0, 0] = 20
        self.assertLoss(logits, programs, target, **kwargs)

    def test_leaf_program_logit(self):
        self.assertUniformLoss([["(1)"]], [np.log(3)])
        self.assertUniformLoss([["(2)"]], [np.log(3)])

    def test_multiple_program_logit(self):
        self.assertUniformLoss([["(1)", "(2)"]], [np.log(3) * 2])
        self.assertUniformLoss([["(1)", "(1)"]], [np.log(3) * 2])

    def test_separate_program_logit(self):
        self.assertUniformLoss([["(1)"], ["(2)"]], [np.log(3)] * 2)
        self.assertUniformLoss([["(1)"], ["(1)"]], [np.log(3)] * 2)

    def test_nested_program_logit(self):
        self.assertUniformLoss([["(+ (1) (2))"]], [np.log(3) * 3])

    def test_nonuniform_loss(self):
        logits = torch.zeros((2, 4, 2, 4))
        logits[0, 0, 0, 1] = np.log(3)  # 1/2 chance of +
        logits[0, 0, 0, 2] = np.log(2)  # 1/3 chance of 1, 1/6 chance of 2
        self.assertLoss(logits, [["(1)"], ["(1)"]], [-np.log(1 / 3), -np.log(1 / 3)])
        self.assertLoss(logits, [["(2)"], ["(2)"]], [-np.log(1 / 6), -np.log(1 / 3)])
        self.assertLoss(
            logits,
            [["(+ (1) (2))"], ["(+ (1) (2))"]],
            [-(np.log(1 / 2) + np.log(1 / 3) * 2), -3 * np.log(1 / 3)],
        )

    def test_variables_loss(self):
        logits = torch.zeros((1, 10, 2, 10))
        self.assertLoss(logits, [["(1)"]], [np.log(4)], family=fam_with_vars)
        self.assertLoss(
            logits,
            [["(call (lam ($0_0)) (1))"]],
            [np.log(4 * 5 * 4)],
            family=fam_with_vars,
        )


class BigramLikelihoodTest(unittest.TestCase):
    def assertLikelihood(self, dist, program, str_prob, family=fam):
        likelihood = family.compute_likelihood(dist, ns.parse_s_expression(program))
        prob = np.exp(likelihood)
        prob = Fraction.from_float(float(prob)).limit_denominator(1000)
        result = f"log({prob})"
        print(result)
        self.assertEqual(result, str_prob)

    def test_leaf(self):
        self.assertLikelihood(fam.uniform(), "(1)", "log(1/3)")

    def test_plus(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (2))", "log(1/27)")

    def test_plus_nested(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (+ (1) (2)))", "log(1/243)")

    def test_leaf_with_variables(self):
        self.assertLikelihood(
            fam_with_vars.uniform(), "(1)", "log(1/4)", family=fam_with_vars
        )

    def test_call_with_variables(self):
        # 1/4 for call
        # 1 for lam
        # 1/5 for $0_0
        # 1/4 for 1
        self.assertLikelihood(
            fam_with_vars.uniform(),
            "(call (lam ($0_0)) (1))",
            "log(1/80)",
            family=fam_with_vars,
        )



from neurosym.program_dist.tree_distribution.preorder_mask.collect_preorder_symbols import (
    annotate_with_alternate_symbols,
)
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    NoopPreorderMask,
    PreorderMask,
)

from . import compression, datasets, examples, search
from .dsl.dsl_factory import DSLFactory
from .program_dist.bigram import (
    BigramProgramCounts,
    BigramProgramCountsBatch,
    BigramProgramDistributionFamily,
)
from .program_dist.tree_distribution.preorder_mask.type_preorder_mask import (
    TypePreorderMask,
)
from .program_dist.tree_distribution.tree_dist_enumerator import enumerate_tree_dist
from .program_dist.tree_distribution.tree_distribution import TreeDistribution
from .programs.s_expression import SExpression
from .programs.s_expression_render import (
    parse_s_expression,
    render_s_expression,
    symbols_for_program,
)
from .search_graph.dsl_search_graph import DSLSearchGraph
from .search_graph.hole_set_chooser import ChooseFirst
from .search_graph.metadata_computer import NoMetadataComputer
from .types.type import (
    ArrowType,
    AtomicType,
    ListType,
    TensorType,
    Type,
    TypeVariable,
    UnificationError,
)
from .types.type_signature import bottom_up_enumerate_types, expansions
from .types.type_string_repr import TypeDefiner, lex, parse_type, render_type
from .types.type_with_environment import Environment, TypeWithEnvironment



from dataclasses import dataclass
from typing import Dict, Generic, List, Set, Tuple, TypeVar

K, V = TypeVar("K"), TypeVar("V")


@dataclass
class TreeTrie(Generic[K, V]):
    """
    Like a Trie, but allows lookup of trees
    """

    leaves: Dict[K, List[V]]
    wild: List[V]
    # trie_children[k] is a tree trie for each child of k
    trie_children: Dict[K, Tuple["TreeTrie[K, V]"]]

    @classmethod
    def empty(cls):
        return cls({}, [], {})

    def insert(self, key, value, *, is_wildcard_predicate):
        children = list(key.children())
        summary = key.node_summary(), len(children)
        if is_wildcard_predicate(key):
            self.wild.append(value)
        if not children:
            self.leaves[summary] = self.leaves.get(summary, []) + [value]
            return
        if summary not in self.trie_children:
            self.trie_children[summary] = tuple(TreeTrie.empty() for _ in children)
        trie_children = self.trie_children[summary]
        for child, trie_child in zip(children, trie_children):
            trie_child.insert(child, value, is_wildcard_predicate=is_wildcard_predicate)

    def query(self, key) -> Set[V]:
        children = list(key.children())
        summary = key.node_summary(), len(children)
        as_leaf = set() if summary not in self.leaves else set(self.leaves[summary])
        as_wild = set(self.wild)
        if summary in self.trie_children:
            as_child = None
            trie_children = self.trie_children[summary]
            assert len(trie_children) == len(children)
            for child, trie_child in zip(children, trie_children):
                query_finding = trie_child.query(child)
                if as_child is None:
                    as_child = query_finding
                else:
                    as_child = as_child & query_finding
                if as_child is not None and not as_child:
                    break
        else:
            as_child = set()
        return as_leaf | as_wild | as_child



import itertools
from abc import abstractmethod

from .search_graph import SearchGraph


class FilterEdgesGraph(SearchGraph):
    def __init__(self, graph):
        self.graph = graph

    @abstractmethod
    def include_edge(self, s, t) -> bool:
        pass

    def initial_node(self):
        return self.graph.initial_node()

    def expand_node(self, node):
        return [t for t in self.graph.expand_node(node) if self.include_edge(node, t)]

    def is_goal_node(self, node):
        return self.graph.is_goal_node(node)


class LimitEdgesGraph(SearchGraph):
    def __init__(self, graph, limit):
        self.graph = graph
        self.limit = limit

    def initial_node(self):
        return self.graph.initial_node()

    def expand_node(self, node):
        return itertools.islice(self.graph.expand_node(node), self.limit)

    def is_goal_node(self, node):
        return self.graph.is_goal_node(node)



from dataclasses import dataclass
from typing import List, Tuple

import torch
from torch import nn

from .base import BaseConfig


@dataclass
class RNNConfig(BaseConfig):
    input_size: int
    hidden_size: int
    output_size: int


class RNN(nn.Module):
    """Abstract RNN module."""

    def __init__(self, config: RNNConfig):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.rnn = nn.RNN(
            config.input_size,
            config.hidden_size,
            batch_first=True,
            num_layers=1,
            bidirectional=False,
            dropout=0.0,
            nonlinearity="tanh",
            bias=True,
        )
        self.fc = nn.Linear(config.hidden_size, config.output_size)

    def seq2class(self, x, hidden: torch.Tensor = None):
        """
        :param x : (batch_size, seq_length, input_size)
        :return out : (batch_size, output_size)
        """
        b, _, _ = x.shape
        h0 = (
            torch.zeros(1, b, self.hidden_size, device=x.device)
            if hidden is None
            else hidden
        )
        out, _ = self.rnn(x, h0)
        out = self.fc(out[:, -1, :])
        return out

    def seq2seq(self, x, hidden: torch.Tensor = None):
        """
        :param x : (batch_size, seq_length, input_size)
        :return out : (batch_size, seq_length, output_size)
        """
        b, s, _ = x.shape
        h0 = (
            torch.zeros(1, b, self.hidden_size, device=x.device)
            if hidden is None
            else hidden
        )
        out, _ = self.rnn(x, h0)
        out = self.fc(out.contiguous().view(b * s, -1)).view(b, s, -1)
        return out

    def forward(self, inp, hidden: torch.Tensor = None):
        pass


class Seq2SeqRNN(RNN):
    def forward(self, inp: torch.Tensor, hidden: torch.Tensor = None):
        return self.seq2seq(inp, hidden)


class Seq2ClassRNN(RNN):
    def forward(self, inp: torch.Tensor, hidden: torch.Tensor = None):
        return self.seq2class(inp, hidden)


def rnn_factory_seq2seq(**kwargs):
    """
    Allows instantiating an MLP module with a given input and output size.
    """

    def construct_model(input_shape: List[Tuple[int]], output_shape: Tuple[int]):
        assert len(input_shape) == 1, "MLP takes a single input only."
        input_size = input_shape[0][-1]
        output_size = output_shape[-1]
        cfg = RNNConfig(
            model_name="rnn",
            input_size=input_size,
            output_size=output_size,
            **kwargs,
        )
        return Seq2SeqRNN(cfg)

    return construct_model


def rnn_factory_seq2class(**kwargs):
    """
    Allows instantiating an MLP module with a given input and output size.
    """

    def construct_model(input_shape: List[Tuple[int]], output_shape: Tuple[int]):
        assert len(input_shape) == 1, "MLP takes a single input only."
        input_size = input_shape[0][-1]
        output_size = output_shape[-1]
        cfg = RNNConfig(
            model_name="rnn",
            input_size=input_size,
            output_size=output_size,
            **kwargs,
        )
        return Seq2ClassRNN(cfg)

    return construct_model



import io
import os

import numpy as np
import pytorch_lightning as pl
import requests
import torch
from permacache import permacache


def get_raw_url(github_folder, filename):
    """
    Get the raw url for a file in a github folder.

    Parameters
    ----------
    github_folder : str
        The github folder containing the file. As a url.
        E.g., github_folder = https://github.com/trishullab/near/tree/master/near_code/data/example
    filename : str
        The name of the file.
            E.g., filename = test_ex_data.npy

    Returns
    -------
    raw_url : str
        The raw url for the file.
        E.g., https://github.com/trishullab/near/raw/master/near_code/data/example/test_ex_data.npy
    """
    raw_url = github_folder.replace("tree", "raw") + "/" + filename
    return raw_url


@permacache("neurosym/data/load_data/load_npy")
def load_npy(path_or_url):
    """
    Load a numpy file from a path or url.

    Parameters
    ----------
    path_or_url : str
        The path or url of the numpy file.

    Returns
    -------
    data : np.ndarray
        The data in the numpy file.
    """
    # pylint: disable=missing-timeout
    if os.path.exists(path_or_url):
        # Load from local path
        data = np.load(path_or_url)
    else:
        data = requests.get(path_or_url).content
        data = np.load(io.BytesIO(data))
    return data


class DatasetFromNpy(torch.utils.data.Dataset):
    """
    A dataset from an array, loaded from a url.

    TODO test/val split
    """

    def __init__(self, input_url, output_url, seed):
        """
        Parameters
        ----------
        url : str
            The url of the numpy file.
        """
        self.inputs = load_npy(input_url)
        self.outputs = load_npy(output_url)
        assert len(self.inputs) == len(self.outputs)
        if seed is not None:
            self.ordering = np.random.RandomState(seed=seed).permutation(
                len(self.inputs)
            )
        else:
            self.ordering = np.arange(len(self.inputs))

    def get_io_dims(self):
        return self.inputs.shape[-1], len(np.unique(self.outputs))

    def __len__(self):
        return len(self.inputs)

    def __getitem__(self, idx):
        return dict(
            inputs=self.inputs[self.ordering[idx]],
            outputs=self.outputs[self.ordering[idx]],
        )


class DatasetWrapper(pl.LightningDataModule):
    def __init__(
        self,
        train: torch.utils.data.Dataset,
        test: torch.utils.data.Dataset,
        batch_size: int = 32,
    ):
        super().__init__()
        self.train = train
        self.test = test
        self.batch_size = batch_size

    def train_dataloader(self):
        return torch.utils.data.DataLoader(self.train, batch_size=self.batch_size)

    def val_dataloader(self):
        return torch.utils.data.DataLoader(self.test, batch_size=self.batch_size)

    def test_dataloader(self):
        return torch.utils.data.DataLoader(self.test, batch_size=self.batch_size)


def numpy_dataset_from_github(
    github_url,
    train_input_path,
    train_output_path,
    test_input_path,
    test_output_path,
):
    """
    Load a dataset from a github url.

    Parameters
    ----------
    github_url : str
        The url of the github folder containing the data.
    train_input_path : str
        The path to the training input data.
    train_output_path : str
        The path to the training output data.
    test_input_path : str
        The path to the test input data.
    test_output_path : str
        The path to the test output data.

    Returns
    -------
    dataset : function seed -> DatasetWrapper
        The dataset, as a function of the seed.
    """
    return lambda train_seed: DatasetWrapper(
        DatasetFromNpy(
            get_raw_url(github_url, train_input_path),
            get_raw_url(github_url, train_output_path),
            train_seed,
        ),
        DatasetFromNpy(
            get_raw_url(github_url, test_input_path),
            get_raw_url(github_url, test_output_path),
            None,
        ),
    )



import sys
from dataclasses import dataclass
from typing import Tuple

import pytorch_lightning as pl
import torch
from torch import nn


@dataclass
class BaseTrainerConfig:
    lr: float = 1e-4
    weight_decay: float = 0.0
    n_epochs: int = 10
    train_steps: int = -1  # Set programmaticaly
    seed: int = 44
    evaluate: bool = True
    resume: str = ""
    scheduler: str = "cosine"
    sav_dir: str = "data/shapeworldonly_checkpoints"
    _filter_param_list: Tuple[str] = ()
    scheduler: str = "cosine"
    optimizer: str = "adam"


class BaseTrainer(pl.LightningModule):
    """
    An abstract class that defines the supporting code to
    train a neural module. We use pytorch-lightning as the
    base framework.
    """

    @classmethod
    def from_arch(cls, model: nn.Module, config: BaseTrainerConfig):
        """
        Instantiates a neural module from a torch.nn.Module.
        """
        return cls(model, config)

    def __init__(self, model: nn.Module, config: BaseTrainerConfig):
        super().__init__()
        self.config = config
        self.model = model
        self.save_hyperparameters()

    def loss(self) -> dict:
        raise NotImplementedError("Loss function not implemented.")

    def _step(self) -> dict:
        raise NotImplementedError("Step function not implemented.")

    def training_step(self, train_batch, batch_idx):
        # pylint: disable=arguments-differ
        del batch_idx
        losses = self._step(validation=False, **train_batch)
        train_loss = sum(losses.values()) / len(losses)
        for key, value in losses.items():
            self.log(f"train_{key}", value.item(), prog_bar=True, sync_dist=True)

        self.log("train_loss", train_loss, prog_bar=True, sync_dist=True)
        return {"loss": train_loss}

    def evaluation_step(self, evaluation_batch, batch_idx, split):
        # pylint: disable=arguments-differ
        del batch_idx
        losses = self._step(validation=True, **evaluation_batch)
        evaluation_loss = sum(losses.values()) / len(losses)
        for key, value in losses.items():
            self.log(f"{split}_{key}", value.item(), prog_bar=True, sync_dist=True)

        self.log(f"{split}_loss", evaluation_loss, prog_bar=True, sync_dist=True)
        return {f"{split}_loss": evaluation_loss}

    def validation_step(self, val_batch, batch_idx):
        # pylint: disable=arguments-differ
        return self.evaluation_step(val_batch, batch_idx, "val")

    def test_step(self, test_batch, batch_idx):
        # pylint: disable=arguments-differ
        return self.evaluation_step(test_batch, batch_idx, "test")

    @staticmethod
    def filter_parameters(named_parameters: dict, filter_param_list: list):
        params = []
        for name, param in named_parameters:
            valid_name = not any(x in name for x in filter_param_list)
            if param.requires_grad and valid_name:
                params.append(param)
            else:
                print(
                    f"WARNING: Parameter {name} removed from optimizer", file=sys.stderr
                )
        return params

    @staticmethod
    def warm_and_decay_lr_scheduler(warmup_steps_pct, decay_steps_pct, total_steps):
        def f(step):
            warmup_steps = warmup_steps_pct * total_steps
            decay_steps = decay_steps_pct * total_steps
            assert step < (
                total_steps + 1
            ), f"Step {step} is greater thantotal steps {total_steps}"
            if step < warmup_steps:
                factor = step / warmup_steps
            else:
                factor = 1
            factor *= 0.5 ** (step / decay_steps)
            return factor

        return f

    def step_lr_scheduler(self, total_steps: int):
        def f(step: int):
            if step < total_steps * 0.3:
                factor = 1
            elif step < total_steps * 0.3:
                factor = 0.1
            else:
                factor = 0.01
            return factor

        return f

    def configure_optimizers(self):
        """
        A rather verbose function that instantiates the optimizer and scheduler.
        """
        # pylint: disable=protected-access
        params = self.filter_parameters(
            self.named_parameters(), self.config._filter_param_list
        )

        match self.config.optimizer:
            case "adam":
                optimizer = torch.optim.Adam(
                    params, lr=self.config.lr, weight_decay=self.config.weight_decay
                )
            case "sgd":
                optimizer = torch.optim.SGD(
                    params, lr=self.config.lr, weight_decay=self.config.weight_decay
                )
            case "adamw":
                optimizer = torch.optim.AdamW(
                    params, lr=self.config.lr, weight_decay=self.config.weight_decay
                )
            case _:
                raise NotImplementedError(
                    f"Optimizer {self.config.optimizer} not implemented"
                )  # noqa: E501

        assert self.config.train_steps != -1, "Train steps not set"
        total_steps = int(self.config.n_epochs * (self.config.train_steps))

        match self.config.scheduler:
            case "none":
                return optimizer
            case "cosine":
                warmup_steps_pct = 0.02
                decay_steps_pct = 0.2
                scheduler = torch.optim.lr_scheduler.LambdaLR(
                    optimizer=optimizer,
                    lr_lambda=self.warm_and_decay_lr_scheduler(
                        warmup_steps_pct, decay_steps_pct, total_steps
                    ),
                )
                return (
                    [optimizer],
                    [
                        {
                            "scheduler": scheduler,
                            "interval": "step",
                        }
                    ],
                )
            case "step":
                scheduler = torch.optim.lr_scheduler.LambdaLR(
                    optimizer=optimizer, lr_lambda=self.step_lr_scheduler(total_steps)
                )
                return (
                    [optimizer],
                    [
                        {
                            "scheduler": scheduler,
                            "interval": "step",
                        }
                    ],
                )
            case _:
                raise ValueError(f"Invalid scheduler {self.config.scheduler}")



from collections import defaultdict
from dataclasses import dataclass
from types import NoneType
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import torch

from neurosym.dsl.dsl import DSL
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    ConjunctionPreorderMask,
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.preorder_mask.type_preorder_mask import (
    TypePreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression
from neurosym.types.type import Type

from .tree_distribution.tree_distribution import TreeProgramDistributionFamily


@dataclass
class BigramProgramDistribution:
    distribution: np.ndarray

    def __post_init__(self):
        assert self.distribution.ndim == 3
        assert self.distribution.shape[0] == self.distribution.shape[2]


@dataclass
class BigramProgramDistributionBatch:
    distribution_batch: np.ndarray

    def __post_init__(self):
        assert isinstance(self.distribution_batch, np.ndarray), type(
            self.distribution_batch
        )
        assert self.distribution_batch.ndim == 4
        assert self.distribution_batch.shape[1] == self.distribution_batch.shape[3]

    def __getitem__(self, i):
        return BigramProgramDistribution(self.distribution_batch[i])


@dataclass
class BigramProgramCounts:
    # map from (parent_sym, parent_child_idx) to map from child_sym to count
    numerators: Dict[Tuple[Tuple[int, int], ...], Dict[int, int]]
    # map from (parent_sym, parent_child_idx) to map from potential child_sym values to count
    denominators: Dict[Tuple[Tuple[int, int], ...], Dict[Tuple[int, ...], int]]

    def add_to_numerator_array(self, arr, batch_idx):
        for [(parent_sym, parent_child_idx)], children in self.numerators.items():
            for child_sym, count in children.items():
                arr[batch_idx, parent_sym, parent_child_idx, child_sym] = count
        return arr

    def add_to_denominator_array(self, arr, batch_idx, backmap):
        for [(parent_sym, parent_child_idx)], children in self.denominators.items():
            for child_syms, count in children.items():
                key = batch_idx, parent_sym, parent_child_idx, backmap[child_syms]
                arr[key] = count
        return arr


@dataclass
class BigramProgramCountsBatch:
    counts: List[BigramProgramCounts]

    def numerators(self, num_symbols, max_arity):
        numerators = np.zeros(
            (len(self.counts), num_symbols, max_arity, num_symbols), dtype=np.int32
        )
        for i, dist in enumerate(self.counts):
            dist.add_to_numerator_array(numerators, i)
        return numerators

    def denominators(self, num_symbols, max_arity):
        denominator_keys = {
            key
            for counts in self.counts
            for mapping in counts.denominators.values()
            for key in mapping.keys()
        }
        denominator_keys = sorted(denominator_keys)
        denominator_keys_backmap = {key: i for i, key in enumerate(denominator_keys)}
        denominators = np.zeros(
            (len(self.counts), num_symbols, max_arity, len(denominator_keys)),
            dtype=np.int32,
        )
        for i, counts in enumerate(self.counts):
            counts.add_to_denominator_array(denominators, i, denominator_keys_backmap)

        return denominators, denominator_keys

    def to_distribution(self, num_symbols, max_arity):
        numerators = self.numerators(num_symbols, max_arity)

        # We do not need to handle denominators here, as this is just
        # a simple conversion from counts to probabilities, and we do
        # not need to handle the case where the denominator is 0.

        return BigramProgramDistributionBatch(counts_to_probabilities(numerators))


class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
    def __init__(
        self,
        dsl: DSL,
        valid_root_types: Union[NoneType, List[Type]] = None,
        *,
        additional_preorder_masks: Tuple[
            Callable[[DSL, TreeDistribution], PreorderMask]
        ] = (),
    ):
        if valid_root_types is not None:
            dsl = dsl.with_valid_root_types(valid_root_types)
        self._dsl = dsl
        self._symbols, self._arities, self._valid_mask = bigram_mask(dsl)
        self._max_arity = max(self._arities)
        self._symbol_to_idx = {sym: i for i, sym in enumerate(self._symbols)}
        self._additional_preorder_masks = additional_preorder_masks

    def underlying_dsl(self) -> DSL:
        return self._dsl

    def parameters_shape(self) -> List[int]:
        return self._valid_mask.shape

    def normalize_parameters(
        self, parameters: torch.Tensor, *, logits: bool, neg_inf=-float("inf")
    ) -> torch.Tensor:
        parameters = parameters.clone()
        mask = torch.tensor(self._valid_mask, device=parameters.device)[None].repeat(
            parameters.shape[0], 1, 1, 1
        )
        parameters[~mask] = -float("inf")
        if logits:
            parameters = parameters.log_softmax(-1)
            parameters[~mask] = neg_inf
        else:
            parameters = parameters.softmax(-1)
            parameters[~mask] = 0
        return parameters

    def with_parameters(
        self, parameters: torch.Tensor
    ) -> BigramProgramDistributionBatch:
        assert (
            parameters.shape[1:] == self.parameters_shape()
        ), f"Expected {self.parameters_shape()}, got {parameters.shape[1:]}"
        parameters = self.normalize_parameters(parameters, logits=False)
        return BigramProgramDistributionBatch(parameters.detach().cpu().numpy())

    def count_programs(self, data: List[List[SExpression]]) -> BigramProgramCountsBatch:
        tree_dist = self.tree_distribution_skeleton
        all_counts = []
        for programs in data:
            numerators, denominators = count_programs(tree_dist, programs)
            all_counts.append(
                BigramProgramCounts(numerators=numerators, denominators=denominators)
            )
        return BigramProgramCountsBatch(all_counts)

    def counts_to_distribution(
        self, counts: BigramProgramCountsBatch
    ) -> BigramProgramDistribution:
        return counts.to_distribution(len(self._symbols), self._max_arity)

    def parameter_difference_loss(
        self, parameters: torch.tensor, actual: BigramProgramCountsBatch
    ) -> torch.float32:
        """
        Let p be a program, g be an ngram, s be a symbol, and d be the `denominator`
            (i.e., the set of possible children for a given parent and position).

        sum_p log P(p | theta)
            = sum_p sum_{(g, s, d) in p} log P(s | g, theta, d)
            = sum_p sum_{(g, s, d) in p} log (exp(theta_{g, s}) / sum_{s' in d} exp(theta_{g, s'}))
            = sum_p sum_{(g, s, d) in p} (theta_{g, s} - log sum_{s' in d} exp(theta_{g, s'}))
            = [sum_p sum_{(g, s, d) in p} theta_{g, s}]
                - [sum_p sum_{(g, s, d) in p} log sum_{s' in d} exp(theta_{g, s'})]
            = [numer] - [denom]

        numer
            = sum_p sum_{(g, s, d) in p} theta_{g, s}
            = sum_g sum_s numcount_{g, s} theta_{g, s}
            = (numcount * theta).sum()

        denom
            = sum_p sum_{(g, s, d) in p} log sum_{s' in d} exp(theta_{g, s'})
            = sum_g sum_d dencount_{g, d} log sum_{s' in d} exp(theta_{g, s'})

        theta_by_denom(g, s', d) = theta_{g, s'} if s' in d, else -inf

        denom
            = sum_g sum_d dencount_{g, d} log sum_{s'} exp(theta_by_denom{g, s', d})

        agg_theta_by_denom(g, d) = logsumexp(theta_by_denom(g, *, d))

        denom
            = sum_g sum_d dencount_{g, d} agg_theta_by_denom(g, d)
            = (dencount * agg_theta_by_denom).sum()

        """

        assert parameters.shape[1:] == self.parameters_shape()
        assert len(parameters.shape) == 4

        numcount = actual.numerators(len(self._symbols), self._max_arity)
        dencount, den_keys = actual.denominators(len(self._symbols), self._max_arity)
        numcount, dencount = [
            torch.tensor(x, device=parameters.device, dtype=torch.float32)
            for x in (numcount, dencount)
        ]

        def agg_across_all_but_batch_axis(x, fn):
            x = x.reshape(x.shape[0], -1)
            return fn(x, -1)

        numer = agg_across_all_but_batch_axis(numcount * parameters, torch.sum)

        theta_by_denom = parameters[..., None].repeat(1, 1, 1, 1, len(den_keys))
        for i, key in enumerate(den_keys):
            mask = torch.ones(
                len(self._symbols), dtype=torch.bool, device=parameters.device
            )
            mask[list(key)] = False
            theta_by_denom[..., mask, i] = -float("inf")
        agg_theta_by_denom = torch.logsumexp(theta_by_denom, dim=-2)
        denom = agg_across_all_but_batch_axis(dencount * agg_theta_by_denom, torch.sum)
        return -(numer - denom)

    def uniform(self):
        return BigramProgramDistribution(counts_to_probabilities(self._valid_mask))

    def compute_tree_distribution(
        self, distribution: Union[BigramProgramDistribution, NoneType]
    ) -> TreeDistribution:
        if isinstance(distribution, BigramProgramDistribution):
            assert isinstance(distribution, BigramProgramDistribution), type(
                distribution
            )
            dist = defaultdict(list)
            for parent, position, child in zip(
                *np.where(distribution.distribution > 0)
            ):
                dist[(parent, position),].append(
                    (child, np.log(distribution.distribution[parent, position, child]))
                )
            dist = {k: sorted(v, key=lambda x: -x[1]) for k, v in dist.items()}
        else:
            assert distribution is None
            dist = None

        return TreeDistribution(
            1,
            dist,
            list(zip(self._symbols, self._arities)),
            lambda tree_dist: ConjunctionPreorderMask(
                tree_dist,
                [TypePreorderMask(tree_dist, self._dsl)]
                + [
                    mask(tree_dist, self._dsl)
                    for mask in self._additional_preorder_masks
                ],
            ),
        )


def bigram_mask(dsl):
    symbols = ["<root>"] + sorted([x.symbol() for x in dsl.productions])

    valid_root_types = dsl.valid_root_types
    symbol_to_idx = {sym: i for i, sym in enumerate(symbols)}
    rules_for = dsl.all_rules(
        care_about_variables=False, valid_root_types=valid_root_types
    )
    arities = [None] * len(symbols)
    arities[0] = 1
    for _, rules in rules_for.items():
        for sym, children in rules:
            assert arities[symbol_to_idx[sym]] in (None, len(children)), str(
                symbol_to_idx
            )
            arities[symbol_to_idx[sym]] = len(children)
    assert all(
        x is not None for x in arities
    ), f"No arities for symbols {[sym for sym, ar in zip(symbols, arities) if ar is None]}"
    valid_mask = np.zeros((len(symbols), max(arities), len(symbols)), dtype=np.bool_)
    root_syms = {sym for t in valid_root_types for sym, _ in rules_for[t]}
    for root_sym in root_syms:
        valid_mask[symbol_to_idx["<root>"], 0, symbol_to_idx[root_sym]] = 1
    for _, rules in rules_for.items():
        for root_sym, child_types in rules:
            for i, ct in enumerate(child_types):
                for child_sym, _ in rules_for.get(ct, ()):
                    valid_mask[symbol_to_idx[root_sym], i, symbol_to_idx[child_sym]] = 1
    return symbols, np.array(arities), valid_mask


def counts_to_probabilities(counts):
    return np.divide(
        counts,
        counts.sum(-1)[..., None],
        out=np.zeros_like(counts, dtype=np.float32),
        where=counts != 0,
    )


def count_programs(tree_dist: TreeDistribution, programs: List[SExpression]):
    """
    Count the productions in the programs, indexed by the path to the node.
    """
    numerators = defaultdict(lambda: defaultdict(int))
    denominators = defaultdict(lambda: defaultdict(int))
    for program in programs:
        preorder_mask = tree_dist.mask_constructor(tree_dist)
        preorder_mask.on_entry(0, 0)
        accumulate_counts(
            tree_dist,
            program,
            numerators,
            denominators,
            ((0, 0),),
            preorder_mask=preorder_mask,
        )
    numerators = {k: dict(v) for k, v in numerators.items()}
    denominators = {k: dict(v) for k, v in denominators.items()}
    return numerators, denominators


def accumulate_counts(
    tree_dist: TreeDistribution,
    program: SExpression,
    numerators: Dict[Tuple[Tuple[int, int], ...], Dict[int, int]],
    denominators: Dict[Tuple[Tuple[int, int], ...], Dict[Tuple[int, ...], int]],
    ancestors: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
):
    parent_position = ancestors[-1][1]
    this_idx = tree_dist.symbol_to_index[program.symbol]
    numerators[ancestors][this_idx] += 1
    possibilities = np.arange(len(tree_dist.symbols))
    mask = preorder_mask.compute_mask(parent_position, possibilities)
    preorder_mask.on_entry(parent_position, this_idx)
    elements = possibilities[mask]
    elements = tuple(int(x) for x in elements)
    denominators[ancestors][elements] += 1
    for j, child in enumerate(program.children):
        new_ancestors = ancestors + ((this_idx, j),)
        new_ancestors = new_ancestors[-tree_dist.limit :]
        accumulate_counts(
            tree_dist,
            child,
            numerators,
            denominators,
            new_ancestors,
            preorder_mask=preorder_mask,
        )
    preorder_mask.on_exit(parent_position, this_idx)



from neurosym.search_graph.dsl_search_graph import DSLSearchGraph
from neurosym.search_graph.hole_set_chooser import ChooseFirst
from neurosym.search_graph.metadata_computer import NoMetadataComputer
from neurosym.search_graph.search_graph_transformer import (
    FilterEdgesGraph,
    LimitEdgesGraph,
)


class FilterUnexpandableNodes(FilterEdgesGraph):
    """
    Represents a search graph that removes nodes that cannot be expanded.
    """

    def include_edge(self, s, t):
        del s
        # TODO(AS) throw out nodes that cannot
        # be expanded because they would be too deep
        return True


def near_graph(dsl, root_type, *, max_num_edges=100, is_goal=lambda x: True):
    graph = DSLSearchGraph(
        dsl,
        root_type,
        ChooseFirst(),
        is_goal,
        NoMetadataComputer(),
    )
    graph = FilterUnexpandableNodes(graph)
    graph = LimitEdgesGraph(graph, max_num_edges)
    return graph



from neurosym.dsl.dsl_factory import DSLFactory
from neurosym.types.type_string_repr import TypeDefiner


def increment_and_return(counter):
    counter[0] += 1
    return counter[0]


t = TypeDefiner()
t.typedef("fn", "(i) -> i")

dslf = DSLFactory()
dslf.typedef("fn", "(i) -> i")
dslf.concrete(
    "x",
    "() -> $fn",
    lambda: lambda t: t,
)

dslf.concrete(
    "1",
    "() -> $fn",
    lambda: lambda t: 1,
)

dslf.concrete(
    "+",
    "($fn, $fn) -> $fn",
    lambda x, y: lambda t: x(t) + y(t),
)

dslf.concrete(
    "*",
    "($fn, $fn) -> $fn",
    lambda x, y: lambda t: x(t) * y(t),
)

dslf.concrete(
    "even?",
    "($fn) -> $fn",
    lambda x: lambda t: x(t) % 2 == 0,
)

dslf.concrete(
    "ite",
    "($fn, $fn, $fn) -> $fn",
    lambda cond, fx, fy: lambda t: fx(t) if cond(t) else fy(t),
)

dslf.parameterized(
    "count",
    "() -> $fn",
    lambda counter: lambda t: increment_and_return(counter),
    dict(counter=lambda: [0]),
)

mutable_arith_combinators_dsl = dslf.finalize()



from dataclasses import dataclass
from typing import Dict, List, Tuple

from frozendict import frozendict

from neurosym.types.type import Type


@dataclass(frozen=True, eq=True)
class Environment:
    """
    Represents a type environment containing the variables.
    """

    _elements: Dict[int, Type]  # actually a frozendict

    @classmethod
    def empty(cls):
        return cls(frozendict())

    def child(self, *new_typs: Tuple[Type]):
        new_typs = new_typs[
            ::-1
        ]  # reverse, so that the first element is the first index
        result = {i + len(new_typs): typ for i, typ in self._elements.items()}
        for i, new_typ in enumerate(new_typs):
            result[i] = new_typ
        return Environment(frozendict(result))

    def parent(self, new_types):
        new_types = new_types[
            ::-1
        ]  # reverse, so that the first element is the first index
        for i, new_typ in enumerate(new_types):
            if i in self._elements:
                assert self._elements[i] == new_typ
        result = {
            i - len(new_types): typ
            for i, typ in self._elements.items()
            if i >= len(new_types)
        }
        return Environment(frozendict(result))

    def merge(self, other: "Environment"):
        """
        Merge two environments.
        """
        result = dict(self._elements)
        # pylint: disable=protected-access
        for i, typ in other._elements.items():
            if i in result:
                assert result[i] == typ
            else:
                result[i] = typ
        return Environment(frozendict(result))

    @classmethod
    def merge_all(cls, *environments: List["Environment"]):
        """
        Merge a list of environments.
        """
        result = {}
        for env in environments:
            # pylint: disable=protected-access
            result.update(env._elements)
        return Environment(frozendict(result))

    def contains_type_at(self, typ: Type, index: int):
        return index in self._elements and self._elements[index] == typ

    def __len__(self):
        return 0 if not self._elements else max(self._elements) + 1

    def short_repr(self):
        return ",".join(
            f"{i}={self._elements[i].short_repr()}" for i in sorted(self._elements)
        )


@dataclass(frozen=True, eq=True)
class PermissiveEnvironmment:
    def child(self, *new_types: Tuple[Type]):
        del new_types
        return self

    def parent(self, new_types):
        del new_types
        return self

    def contains_type_at(self, typ: Type, index: int):
        del typ, index
        return True

    def __len__(self):
        return 0


@dataclass(frozen=True, eq=True)
class TypeWithEnvironment:
    """
    Represents a type in a given environment.
    """

    typ: Type
    env: Environment



from dataclasses import dataclass
from typing import Dict

from ..programs.s_expression import InitializedSExpression, SExpression
from ..programs.s_expression_render import render_s_expression
from ..types.type_signature import TypeSignature
from .production import FunctionLikeProduction


@dataclass
class AbstractionIndexParameter:
    """
    Represents a parameter of an abstraction.
    Used in SExpressions and InitializedSExpressions.
    """

    index: int

    def apply(self, inputs):
        """
        Map the parameter to a value.
        """
        return AbstractionParameter(inputs[self.index])

    def __initialize__(self, dsl):
        del dsl
        return self

    def __to_pair__(self, for_stitch):
        del for_stitch
        return f"#{self.index}"


@dataclass
class AbstractionParameter:
    """
    Represents a parameter of an abstraction, with an initialized value.

    Used in InitializedSExpressions.
    """

    value: object

    def __compute_value__(self, dsl):
        del dsl
        return self.value


@dataclass
class AbstractionProduction(FunctionLikeProduction):
    """
    Represents an abstraction in a DSL.
    """

    _symbol: str
    _type_signature: TypeSignature
    _body: SExpression

    def base_symbol(self):
        return self._symbol

    def get_index(self):
        return None

    def with_index(self, index):
        raise NotImplementedError

    def type_signature(self) -> TypeSignature:
        return self._type_signature

    def initialize(self, dsl) -> Dict[str, object]:
        return dict(
            body=dsl.initialize(self._body),
        )

    def evaluate(self, dsl, state, inputs):
        initialized_body = state["body"]
        return dsl.compute(with_index_parameters(initialized_body, inputs))

    def render_as_lambda(self):
        body = render_s_expression(self._body)
        arguments = " ".join(
            render_s_expression(AbstractionIndexParameter(i))
            for i in range(self._type_signature.arity())
        )
        return f"(lam-abstr ({arguments}) {body})"

    def render(self):
        return f"{self._symbol:>15} :: {self._type_signature.render()} = {self.render_as_lambda()}"


def with_index_parameters(ise: InitializedSExpression, inputs: tuple):
    """
    With the given inputs, replace all AbstractionIndexParameters in the given
    InitializedSExpression with AbstractionParameters.

    Args:
        ise: The InitializedSExpression.
        inputs: The inputs.

    Returns a new InitializedSExpression.
    """
    if isinstance(ise, AbstractionIndexParameter):
        return ise.apply(inputs)
    return InitializedSExpression(
        ise.symbol,
        tuple(with_index_parameters(x, inputs) for x in ise.children),
        ise.state,
    )



"""
Test dsl/sequentual_dsl.py with NEAR search graph.

We conduct the following tests:
- Sanity check: We can find a program.
- BFS: We can find a program with BFS.
- Astar: We can find a program with bounded Astar search.
- Enumerate: We can enumerate all programs of a certain size.
- Full Integration test: Can we run a full iteration of NEAR.
NEAR Integration tests.
"""

import unittest

import pytest
import torch

import neurosym as ns
from neurosym.examples import near

from .utils import assertDSLEnumerable


class TestNEARSequentialDSL(unittest.TestCase):
    def test_sequential_dsl_astar(self):
        """
        A minimal implementation of NEAR with a simple DSL.
        search = A-star
        heuristic = validation score after training for N epochs. (pl.Trainer)
        goal = Fully symbolic program. (handled in: search_graph/dsl_search_graph.py)
        test_predicate = score on testing set (pl.Trainer)
        """
        datamodule = ns.datasets.near_data_example(train_seed=0)
        input_dim, output_dim = datamodule.train.get_io_dims()
        original_dsl = near.example_rnn_dsl(input_dim, output_dim)
        trainer_cfg = near.NEARTrainerConfig(
            max_seq_len=100,
            n_epochs=10,
            num_labels=output_dim,
            train_steps=len(datamodule.train),
        )
        t = ns.TypeDefiner(L=input_dim, O=output_dim)
        t.typedef("fL", "{f, $L}")
        t.typedef("fO", "{f, $O}")
        neural_dsl = near.NeuralDSL.from_dsl(
            dsl=original_dsl,
            modules={
                **near.create_modules(
                    "mlp",
                    [t("($fL) -> $fL"), t("($fL) -> $fO")],
                    near.mlp_factory(hidden_size=10),
                ),
                **near.create_modules(
                    "rnn_seq2seq",
                    [t("([$fL]) -> [$fL]"), t("([$fL]) -> [$fO]")],
                    near.rnn_factory_seq2seq(hidden_size=10),
                ),
                **near.create_modules(
                    "rnn_seq2class",
                    [t("([$fL]) -> $fL"), t("([$fL]) -> $fO")],
                    near.rnn_factory_seq2class(hidden_size=10),
                ),
            },
        )

        def validation_cost(node):
            import pytorch_lightning as pl

            trainer = pl.Trainer(
                max_epochs=10,
                devices="auto",
                accelerator="cpu",
                enable_checkpointing=False,
                logger=False,
                callbacks=[],
            )
            try:
                initialized_p = neural_dsl.initialize(node.program)
            except near.PartialProgramNotFoundError:
                return 10000

            model = neural_dsl.compute(initialized_p)
            if not isinstance(model, torch.nn.Module):
                del model
                del initialized_p
                model = near.TorchProgramModule(dsl=neural_dsl, program=node.program)
            pl_model = near.NEARTrainer(model, config=trainer_cfg)
            trainer.fit(
                pl_model, datamodule.train_dataloader(), datamodule.val_dataloader()
            )
            return trainer.callback_metrics["val_loss"].item()

        def checker(node):
            """
            In NEAR, any program that has no holes is valid.
            The hole checking is done before this function will
            be called so we can assume that the program has no holes.
            """
            return (
                set(ns.symbols_for_program(node.program)) - set(original_dsl.symbols())
                == set()
            )

        g = near.near_graph(
            neural_dsl,
            ns.parse_type(
                s="([{f, $L}]) -> [{f, $O}]",
                env=ns.TypeDefiner(L=input_dim, O=output_dim),
            ),
            is_goal=checker,
        )
        # succeed if this raises StopIteration
        with pytest.raises(StopIteration):
            n_iter = 0
            iterator = ns.search.bounded_astar(g, validation_cost, max_depth=3)
            while True:
                print("iteration: ", n_iter)
                n_iter += 1
                node = next(iterator)
                self.assertIsNotNone(node)
                if n_iter > 30:
                    break

    def test_sequential_dsl_enumerate(self):
        """
        Enumerate all programs in dsl upto fixed depth. This test case makes
        sure all DSL combinations upto a fixed depth are valid.
        """
        self.maxDiff = None
        dsl = near.example_rnn_dsl(10, 4)

        assertDSLEnumerable(dsl, "([$fL]) -> [$fO]")



from .load_data import numpy_dataset_from_github

data = numpy_dataset_from_github(
    "https://github.com/trishullab/near/tree/master/near_code/data/example",
    "train_ex_data.npy",
    "train_ex_labels.npy",
    "test_ex_data.npy",
    "test_ex_labels.npy",
)



import unittest

import neurosym as ns

from .bigram_test import dsl_with_vars, fam_with_vars


class TreeDistributionTest(unittest.TestCase):
    def collect_symbols(self, program):
        return ns.render_s_expression(
            ns.annotate_with_alternate_symbols(
                ns.parse_s_expression(program),
                fam_with_vars.tree_distribution_skeleton,
                lambda tree_dist: ns.TypePreorderMask(tree_dist, dsl_with_vars),
            )
        )

    def test_basic_program(self):
        self.assertEqual(
            self.collect_symbols("(+ (1) (1))"),
            "(+/+,1,2,call (1/+,1,2,call) (1/+,1,2,call))",
        )

    def test_program_with_variables(self):
        self.assertEqual(
            self.collect_symbols("(call (lam ($0_0)) (1))"),
            "(call/+,1,2,call (lam/lam ($0_0/$0_0,+,1,2,call)) (1/+,1,2,call))",
        )



import unittest

import neurosym as ns
from neurosym.examples import near

ldsl = ns.examples.dreamcoder.list_dsl("[i] -> i")


class TestListDSL(unittest.TestCase):
    def test_show_dsl(self):
        expected = """
         lam_11 :: L<#body|[i]> -> [i] -> #body
           $0_3 :: V<[i]@0>
        """
        actual = ldsl.render()
        print(actual)
        self.assertTrue(
            {line.strip() for line in expected.strip().split("\n")}.issubset(
                {line.strip() for line in actual.strip().split("\n")}
            )
        )

    def test_basic_dsl(self):
        self.maxDiff = None
        dsl = ldsl

        def is_goal(x):
            try:
                fn = dsl.compute(dsl.initialize(x.program))
                if fn([1, 2, 3]) != 2:
                    return False
                if fn([0, 7, 3, 5]) != 7:
                    return False
                return True
            except:  # pylint: disable=bare-except
                return False

        g = near.near_graph(
            dsl,
            ns.parse_type("[i] -> i"),
            is_goal=is_goal,
        )
        it = ns.search.bfs(g)
        node = next(it).program
        self.assertEqual(
            ns.render_s_expression(node),
            "(lam_11 (index (1) ($0_3)))",
        )



from dataclasses import dataclass
from typing import Callable, Dict, Tuple

from torch import nn

from neurosym.types.type_signature import FunctionTypeSignature

from ...dsl.dsl import DSL
from ...dsl.production import ParameterizedProduction, Production
from ...programs.hole import Hole
from ...programs.s_expression import InitializedSExpression, SExpression
from ...types.type import ArrowType, ListType, TensorType, Type


class PartialProgramNotFoundError(Exception):
    """
    Raised when a partial program cannot be found.
    """


@dataclass
class NeuralDSL(DSL):
    """
    A neural DSL extends `DSL` to handle neural heuristics (ie: type-appropriate NN productions)
    These neural heuristics can be used to fill holes in partial programs.
    Required to run NEAR.
    """

    # partial_programs: Dict[Type, SExpression]
    type_to_symbol: Dict[Type, str]

    @classmethod
    def from_dsl(
        cls, dsl: DSL, modules: Dict[Type, Tuple[str, Callable[[], nn.Module]]]
    ):
        """
        Creates a NeuralDSL from a DSL and a set of type specific modules.

        The type specific modules are used to fill holes in partial programs.

        Args:
            dsl: The DSL to extend.
            modules: A dictionary mapping types to tags and functions that
                are used to initialize the modules for that type.

        Returns:
            A NeuralDSL.
        """
        partial_productions = []
        type_to_symbol = {}

        count_by_tag = {}
        for fn_type, (tag, module_template) in modules.items():
            assert isinstance(
                fn_type, ArrowType
            ), f"Type of partial NN module must be an ArrowType, got {fn_type}"
            count_by_tag[tag] = count_by_tag.get(tag, 0) + 1
            identifier = f"__neural_dsl_internal_{tag}_{count_by_tag[tag]}"
            type_to_symbol[fn_type] = identifier
            # pylint: disable=unexpected-keyword-arg
            module_c_prod = ParameterizedProduction(
                identifier,
                FunctionTypeSignature([], fn_type),
                lambda initialized_module: initialized_module,
                index=None,
                initializers=dict(initialized_module=module_template),
            )

            partial_productions.append(module_c_prod)

        productions = dsl.productions + partial_productions

        return cls(
            productions=productions,
            valid_root_types=dsl.valid_root_types,
            max_type_depth=dsl.max_type_depth,
            max_env_depth=dsl.max_env_depth,
            type_to_symbol=type_to_symbol,
        )

    def get_partial_program(self, hole: Hole) -> Production:
        """
        Returns a production that can be used to fill the given hole.
        """
        return SExpression(
            self.type_to_symbol[hole.twe.typ],
            [],
        )

    def initialize(self, program: SExpression) -> InitializedSExpression:
        """
        Initializes all the productions in the given program.

        Returns a new program with the same structure, but with all the productions
        initialized.
        """
        if isinstance(program, Hole):
            try:
                prog = self.get_partial_program(program)
            except KeyError as e:
                raise PartialProgramNotFoundError(
                    f"Cannot initialize program {program}."
                ) from e
        else:
            prog = program

        if hasattr(prog, "__initialize__"):
            return prog.__initialize__(self)
        prod = self.get_production(prog.symbol)
        return InitializedSExpression(
            prog.symbol,
            tuple(self.initialize(child) for child in prog.children),
            prod.initialize(self),
        )


def create_module_for_type(module_factory, t):
    shape = compute_io_shape(t)
    return lambda: module_factory(*shape)


def create_modules(tag, types, module_factory):
    return {t: (tag, create_module_for_type(module_factory, t)) for t in types}


def compute_io_shape(t):
    """
    t : ArrowType
    returns: dict(input_shape, output_shape)
        input_shape: list of tuples (shape, type)
        output_shape: tuple (shape, type)
    """
    assert isinstance(t, ArrowType)
    input_types = t.input_type
    output_type = t.output_type

    def get_shape(t):
        match t:
            case TensorType(_, shape):
                return shape
            case ListType(element_type):
                return get_shape(element_type)
            case _:
                raise NotImplementedError(f"Cannot compute shape for type {t}")

    input_shape = [get_shape(t) for t in input_types]
    output_shape = get_shape(output_type)
    return input_shape, output_shape



import unittest

import neurosym as ns

from ..utils import assertDSL


class TestDuplicateProduction(unittest.TestCase):
    def test_basic_duplicate(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("1", "() -> i", lambda x: x)
        self.assertRaisesRegex(
            ValueError,
            "Duplicate declarations for production: 1",
            dslf.finalize,
        )


class TestPruning(unittest.TestCase):
    def test_basic_pruning(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("identity", "i -> i", lambda x: x)
        dslf.concrete("add", "(i, i) -> i", lambda x, y: x + y)
        dslf.concrete("convert", "#x -> f", float)
        dslf.prune_to("f")
        dsl = dslf.finalize()
        assertDSL(
            self,
            dsl.render(),
            """
            1 :: () -> i
            add :: (i, i) -> i
            convert_0 :: f -> f
            convert_1 :: i -> f
            identity :: i -> i
            """,
        )

    def test_pruning_error(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("identity", "i -> i", lambda x: x)
        dslf.concrete("add", "(i, i) -> i", lambda x, y: x + y)
        dslf.concrete("convert", "#x -> i", float)
        dslf.prune_to("f")
        self.assertRaisesRegex(
            TypeError,
            "All productions for .* were pruned. Check that the target types are correct.",
            dslf.finalize,
        )

    def test_pruning_specific(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("identity", "f -> i", lambda x: x)
        dslf.concrete("add", "(i, i) -> i", lambda x, y: x + y)
        dslf.prune_to("i")
        self.assertRaises(
            TypeError,
            "All productions for identity were pruned. Check that the target types are correct.",
            dslf.finalize,
        )

    def test_allow_pruning(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("identity", "f -> i", lambda x: x)
        dslf.concrete("add", "(i, i) -> i", lambda x, y: x + y)
        dslf.prune_to("i", tolerate_pruning_entire_productions=True)
        dsl = dslf.finalize()
        assertDSL(
            self,
            dsl.render(),
            """
            1 :: () -> i
            add :: (i, i) -> i
            """,
        )

    def test_pruning_with_call(self):
        dslf = ns.DSLFactory()
        dslf.concrete("1", "() -> i", lambda x: x)
        dslf.concrete("call", "(i -> i, i) -> i", lambda x, y: x(y))
        dslf.lambdas()
        dslf.prune_to("i")
        dsl = dslf.finalize()
        assertDSL(
            self,
            dsl.render(),
            """
            $0_0 :: V<i@0>
            $1_0 :: V<i@1>
            $2_0 :: V<i@2>
            $3_0 :: V<i@3>
            1 :: () -> i
            call :: (i -> i, i) -> i
            lam :: L<#body|i> -> i -> #body
            """,
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neurosym",
    version="0.0.22",
    author="Kavi Gupta",
    author_email="kavig+neurosym@mit.edu",
    description="Neurosymbolic library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/neurosym-lib",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "frozendict==2.3.8",
        "torch",
        "torchvision",
        "torchaudio",
        "numpy",
        "pytorch-lightning",
        "permacache",
        "requests",
        "stitch-core==0.1.25",
        "scikit-learn",
        "s-exp-parser==1.3.1",
    ],
)



import torch


def ite_torch(condition, if_true, if_else):
    def _ite(*args):
        cond = torch.sigmoid(condition(*args))
        true_val = if_true(*args)
        false_val = if_else(*args)
        if len(cond.shape) == 2 and len(true_val.shape) == 3:
            cond = cond.unsqueeze(1)
        cond = cond.expand_as(true_val)
        return cond * true_val + (1 - cond) * false_val

    return _ite



import unittest

import neurosym as ns


class TestRender(unittest.TestCase):
    def test_atomic(self):
        self.assertEqual(ns.render_type(ns.AtomicType("int")), "int")

    def test_tensor(self):
        self.assertEqual(
            ns.render_type(ns.TensorType(ns.AtomicType("int"), (3, 4))), "{int, 3, 4}"
        )

    def test_list(self):
        self.assertEqual(ns.render_type(ns.ListType(ns.AtomicType("int"))), "[int]")

    def test_arrow(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (ns.AtomicType("int"), ns.AtomicType("float")),
                    ns.AtomicType("bool"),
                )
            ),
            "(int, float) -> bool",
        )

    def test_arrow_single_arg(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool"))
            ),
            "int -> bool",
        )

    def test_arrow_single_arg_list(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (ns.ListType(ns.AtomicType("int")),), ns.AtomicType("bool")
                )
            ),
            "[int] -> bool",
        )

    def test_arrow_single_arg_tensor(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (ns.TensorType(ns.AtomicType("int"), (3, 4)),),
                    ns.AtomicType("bool"),
                )
            ),
            "{int, 3, 4} -> bool",
        )

    def test_arrow_single_arg_arrow(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool")),),
                    ns.AtomicType("bool"),
                )
            ),
            "(int -> bool) -> bool",
        )

    def test_arrow_single_arg_arrow_list(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (
                        ns.ListType(
                            ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool"))
                        ),
                    ),
                    ns.AtomicType("bool"),
                )
            ),
            "[int -> bool] -> bool",
        )

    def test_arrow_returns_arrow(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (ns.AtomicType("int"),),
                    ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool")),
                )
            ),
            "int -> int -> bool",
        )

    def test_no_args(self):
        self.assertEqual(
            ns.render_type(ns.ArrowType((), ns.AtomicType("bool"))),
            "() -> bool",
        )

    def test_nested(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    (
                        ns.TensorType(ns.AtomicType("int"), (3, 4)),
                        ns.ListType(ns.AtomicType("float")),
                    ),
                    ns.AtomicType("bool"),
                )
            ),
            "({int, 3, 4}, [float]) -> bool",
        )

    def test_complex(self):
        self.assertEqual(
            ns.render_type(
                ns.ArrowType(
                    input_type=(
                        ns.TensorType(dtype=ns.AtomicType(name="f"), shape=(10,)),
                    ),
                    output_type=ns.TensorType(
                        dtype=ns.AtomicType(name="f"), shape=(10,)
                    ),
                )
            ),
            "{f, 10} -> {f, 10}",
        )

    def test_variable(self):
        self.assertEqual(ns.render_type(ns.TypeVariable(name="f")), "#f")


class TestLex(unittest.TestCase):
    def test_tensor(self):
        self.assertEqual(ns.lex("{int, 3, 4}"), ["{", "int", ",", "3", ",", "4", "}"])

    def test_list(self):
        self.assertEqual(ns.lex("[int]"), ["[", "int", "]"])

    def test_arrow(self):
        self.assertEqual(
            ns.lex("(int, float) -> bool"),
            ["(", "int", ",", "float", ")", "->", "bool"],
        )

    def test_no_args(self):
        self.assertEqual(ns.lex("() -> bool"), ["(", ")", "->", "bool"])

    def test_dollar(self):
        self.assertEqual(ns.lex("$x"), ["$x"])

    def test_dollar_arrow(self):
        self.assertEqual(ns.lex("$fL -> $fL"), ["$fL", "->", "$fL"])


class TestParse(unittest.TestCase):
    def test_atomic(self):
        self.assertEqual(ns.parse_type("int"), ns.AtomicType("int"))

    def test_tensor(self):
        self.assertEqual(
            ns.parse_type("{int, 3, 4}"), ns.TensorType(ns.AtomicType("int"), (3, 4))
        )

    def test_list(self):
        self.assertEqual(ns.parse_type("[int]"), ns.ListType(ns.AtomicType("int")))

    def test_arrow(self):
        self.assertEqual(
            ns.parse_type("(int, float) -> bool"),
            ns.ArrowType(
                (ns.AtomicType("int"), ns.AtomicType("float")), ns.AtomicType("bool")
            ),
        )

    def test_arrow_single_arg(self):
        self.assertEqual(
            ns.parse_type("int -> bool"),
            ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool")),
        )

    def test_arrow_single_arg_list(self):
        self.assertEqual(
            ns.parse_type("[int] -> bool"),
            ns.ArrowType((ns.ListType(ns.AtomicType("int")),), ns.AtomicType("bool")),
        )

    def test_arrow_returns_arrow(self):
        self.assertEqual(
            ns.parse_type("int -> int -> bool"),
            ns.ArrowType(
                (ns.AtomicType("int"),),
                ns.ArrowType((ns.AtomicType("int"),), ns.AtomicType("bool")),
            ),
        )

    def test_multi_arg_arrow_returns_arrow(self):
        t = "(int, float) -> int -> bool"
        self.assertEqual(
            ns.render_type(ns.parse_type(t)),
            t,
        )

    def test_no_args(self):
        self.assertEqual(
            ns.parse_type("() -> bool"),
            ns.ArrowType((), ns.AtomicType("bool")),
        )

    def test_nested(self):
        self.assertEqual(
            ns.parse_type("({int, 3, 4}, [float]) -> bool"),
            ns.ArrowType(
                (
                    ns.TensorType(ns.AtomicType("int"), (3, 4)),
                    ns.ListType(ns.AtomicType("float")),
                ),
                ns.AtomicType("bool"),
            ),
        )

    def test_single_arg(self):
        self.assertEqual(
            ns.render_type(ns.parse_type("([{f, 10}]) -> {f, 20}")),
            "[{f, 10}] -> {f, 20}",
        )

    def test_unparenthesized_arrow_inside_argument(self):
        t = "(i -> b, [i]) -> [i]"
        self.assertEqual(ns.render_type(ns.parse_type(t)), t)

    def test_bad_parse(self):
        self.assertRaises(Exception, lambda: ns.render_type(ns.parse_type("f -> f]")))

    def test_list_of_arrows(self):
        t = "[i -> i]"
        self.assertEqual(ns.render_type(ns.parse_type(t)), t)



import unittest

import numpy as np
import scipy.stats


class ProbabilityTester(unittest.TestCase):
    def assertBinomial(self, n, p, pdelta, k):
        plow = p - pdelta
        phigh = p + pdelta
        kmin = int(plow * n)
        kmax = int(phigh * n)

        prob_mass = scipy.stats.binom.cdf(kmax, n, p) - scipy.stats.binom.cdf(
            kmin - 1, n, p
        )
        self.assertTrue(kmin > 0, f"Expected kmin > 0, got {kmin}")
        self.assertTrue(kmax < n - 1)
        self.assertTrue(
            prob_mass > 0.99, f"Insufficient probability mass: {prob_mass:.3f}"
        )
        self.assertTrue(kmin <= k <= kmax, f"Expected {k} to be in [{kmin}, {kmax}]")

    def assertSameProbDist(self, dist, target):
        print(dist.distribution.tolist())
        np.testing.assert_almost_equal(dist.distribution, target)



from functools import reduce

from neurosym.dsl.dsl_factory import DSLFactory


def list_dsl(*output_types):
    dslf = DSLFactory(max_overall_depth=5, max_expansion_steps=3)

    for i in range(6):
        dslf.concrete(str(i), "() -> i", lambda i=i: i)

    dslf.concrete("empty", "() -> [#T]", lambda: [])
    dslf.concrete("singleton", "#T -> [#T]", lambda x: [x])
    dslf.concrete("range", "i -> [i]", lambda x: list(range(x)))
    dslf.concrete("++", "([#T], [#T]) -> [#T]", lambda x, y: x + y)
    dslf.concrete(
        "mapi",
        "((i, #T) -> #R, [#T]) -> [#R]",
        lambda f: lambda x: [f(i, x) for i, x in enumerate(x)],
    )
    dslf.concrete(
        "reducei",
        "((i, #R, #T) -> #R, #R, [#T]) -> #R",
        lambda f: lambda x: lambda y: reduce(lambda x, y: f(i, x, y), x, y),
    )

    dslf.concrete("true", "() -> b", lambda: True)
    dslf.concrete("not", "b -> b", lambda x: not x)
    dslf.concrete("and", "(b, b) -> b", lambda x, y: x and y)
    dslf.concrete("or", "(b, b) -> b", lambda x, y: x or y)
    dslf.concrete("i", "(b, #T, #T) -> #T", lambda x, y, z: y if x else z)
    dslf.concrete("sort", "([#T]) -> [#T]", sorted)
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("*", "(i, i) -> i", lambda x, y: x * y)
    dslf.concrete("negate", "i -> i", lambda x: -x)
    dslf.concrete("mod", "(i, i) -> i", lambda x, y: x % y)
    dslf.concrete("eq?", "(i, i) -> b", lambda x, y: x == y)
    dslf.concrete("gt?", "(i, i) -> b", lambda x, y: x > y)
    dslf.concrete(
        "is-prime",
        "i -> b",
        lambda x: x > 1 and all(x % i for i in range(2, min(1 + int(x**0.5), x))),
    )
    dslf.concrete("is-square", "i -> b", lambda x: x > 1 and int(x**0.5) ** 2 == x)
    dslf.concrete("sum", "[i] -> i", sum)
    # # (lambda (lambda (reduce (lambda (lambda (+ $0 $1))) 0 $0)))
    dslf.concrete("reverse", "[#T] -> [#T]", lambda x: x[::-1])
    # (lambda (reduce (lambda (lambda (++ (singleton $0) $1))) empty $0))
    dslf.concrete(
        "all",
        "((#T) -> b, [#T]) -> b",
        lambda f: lambda x: all(f(i) for i in x),
    )
    # (lambda (lambda (reduce (lambda (lambda (and $0 $1))) false (map $1 $0))))
    dslf.concrete(
        "any",
        "((#T) -> b, [#T]) -> b",
        lambda f: lambda x: any(f(i) for i in x),
    )
    # (lambda (lambda (reduce (lambda (lambda (or $0 $1))) true (map $1 $0))))
    dslf.concrete("index", "(i, [#T]) -> #T", lambda x, y: y[x])
    # (lambda (lambda (reducei (lambda (lambda (lambda (if (eq? $1 $4) $0 0)))) 0 $0)))
    dslf.concrete(
        "filter",
        "((#T) -> b, [#T]) -> [#T]",
        lambda f: lambda x: [i for i in x if f(i)],
    )
    # (lambda (lambda
    #   (reduce
    #       (lambda (lambda (++ $1 (if ($3 $0) (singleton $0) empty))))
    #       empty
    #       $0)))
    dslf.concrete("slice", "(i, i, [#T]) -> [#T]", lambda x: lambda y: lambda z: z[x:y])
    # (lambda (lambda (lambda
    #   (reducei
    #     (lambda (lambda (lambda
    #         (++
    #             $2
    #             (if
    #                 (and
    #                     (or (gt? $1 $5) (eq? $1 $5))
    #                     (not (or (gt? $4 $1) (eq? $1 $4))))
    #                 (singleton $0)
    #                 empty)))))
    #     empty
    #     $0))))

    dslf.no_zeroadic()

    dslf.lambdas(max_type_depth=3)
    dslf.prune_to(*output_types, prune_variables=False)

    return dslf.finalize()



from abc import ABC, abstractmethod


class MetadataComputer(ABC):
    """
    Represents a way to compute metadata for a node in a search graph.
    """

    @abstractmethod
    def for_initial_node(self):
        """
        Return the metadata for the initial node in the search graph.
        """

    @abstractmethod
    def for_expanded_node(self, node, expanded_node):
        """
        Return the metadata for the expanded node in the search graph.
        """


class NoMetadataComputer(MetadataComputer):
    """
    Represents a metadata computer that returns empty metadata for all nodes.
    """

    def for_initial_node(self):
        return {}

    def for_expanded_node(self, node, expanded_node):
        return {}



import copy
from typing import Dict, List

import numpy as np

from ..types.type import ArrowType, AtomicType, TypeVariable
from ..types.type_signature import (
    FunctionTypeSignature,
    LambdaTypeSignature,
    VariableTypeSignature,
    expansions,
    signature_expansions,
    type_universe,
)
from ..types.type_string_repr import TypeDefiner
from .dsl import DSL
from .production import (
    ConcreteProduction,
    LambdaProduction,
    ParameterizedProduction,
    Production,
    VariableProduction,
)


class DSLFactory:
    """
    A factory for creating DSLs.

    Example usage:
    ```
    factory = DSLFactory()
    factory.typedef("fn", "(i) -> i")
    factory.concrete("inc", "$fn", lambda x: x + 1)
    factory.concrete("const_0", "$fn", lambda x: 0)
    factory.concrete("compose", "($fn, $fn) -> $fn", lambda f, g: lambda x: f(g(x))
    factory.finalize()
    """

    def __init__(
        self, max_expansion_steps=np.inf, max_env_depth=4, max_overall_depth=6, **env
    ):
        self.t = TypeDefiner(**env)
        self._concrete_productions = []
        self._parameterized_productions = []
        self._signatures = []
        self._known_types = []
        self._no_zeroadic = False
        self.lambda_parameters = None
        self.max_expansion_steps = max_expansion_steps
        self.max_overall_depth = max_overall_depth
        self.max_env_depth = max_env_depth
        self.prune = False
        self.target_types = None
        self.prune_variables = False
        self.tolerate_pruning_entire_productions = False

    def typedef(self, key, type_str):
        """
        Define a type.
        """
        self.t.typedef(key, type_str)

    def filtered_type_variable(self, key, type_filter):
        """
        Define a filtered type variable.
        """
        self.t.filtered_type_variable(key, type_filter)

    def known_types(self, *types):
        """
        Add known types to the DSL.
        """
        self._known_types.extend(self.t(typ) for typ in types)

    def no_zeroadic(self):
        """
        Disable zeroadic types (types with no arguments).
        """
        self._no_zeroadic = True

    def lambdas(self, max_arity=2, max_type_depth=4, max_env_depth=4):
        """
        Define a type.
        """
        self.lambda_parameters = dict(
            max_arity=max_arity,
            max_type_depth=max_type_depth,
            max_env_depth=max_env_depth,
        )

    def concrete(self, symbol, type_str, fn):
        """
        Add a concrete production to the DSL.
        """
        sig = self.t.sig(type_str)
        self._concrete_productions.append(
            (
                symbol,
                sig,
                fn,
            )
        )
        self._signatures.append(sig)

    def parameterized(self, symbol, type_str, fn, parameters):
        """
        Add a parameterized production to the DSL.
        """
        sig = self.t.sig(type_str)
        self._parameterized_productions.append(
            (
                symbol,
                sig,
                fn,
                parameters,
            )
        )
        self._signatures.append(sig)

    def prune_to(
        self,
        *target_types,
        prune_variables=True,
        tolerate_pruning_entire_productions=False,
    ):
        """
        Prune the DSL to only include productions that can be constructed from the given
        target types.
        """
        self.prune = True
        self.target_types = [self.t(x) for x in target_types]
        self.prune_variables = prune_variables
        self.tolerate_pruning_entire_productions = tolerate_pruning_entire_productions

    def _expansions_for_single_production(
        self, terminals, type_constructors, constructor, symbol, sig, *args
    ):
        sigs = list(
            signature_expansions(
                sig,
                terminals,
                type_constructors,
                max_expansion_steps=self.max_expansion_steps,
                max_overall_depth=self.max_overall_depth,
            )
        )
        assert len(sigs) > 0, f"No expansions within depth/step bounds for {symbol}"

        prods = [
            constructor(symbol, FunctionTypeSignature.from_type(expansion), *args)
            for expansion in sigs
        ]

        return {symbol: Production.reindex(prods)}

    def _expansions_for_all_productions(
        self, expand_to, terminals, type_constructors, args
    ):
        result = {}
        for arg in args:
            for_prod = self._expansions_for_single_production(
                expand_to, terminals, type_constructors, *arg
            )
            duplicate_keys = sorted(set(for_prod.keys()) & set(result.keys()))
            if duplicate_keys:
                raise ValueError(
                    f"Duplicate declarations for production: {duplicate_keys[0]}"
                )
            result.update(for_prod)
        return result

    def finalize(self):
        """
        Finalize the DSL.
        """

        known_types = [x.astype() for x in self._signatures] + self._known_types

        universe = type_universe(known_types, no_zeroadic=self._no_zeroadic)

        sym_to_productions: Dict[str, List[Production]] = {}
        sym_to_productions.update(
            self._expansions_for_all_productions(
                *universe, ConcreteProduction, self._concrete_productions
            )
        )
        sym_to_productions.update(
            self._expansions_for_all_productions(
                *universe, ParameterizedProduction, self._parameterized_productions
            )
        )

        stable_symbols = set()

        if self.lambda_parameters is not None:
            types, constructors_lambda = type_universe(
                known_types,
                require_arity_up_to=self.lambda_parameters["max_arity"],
                no_zeroadic=self._no_zeroadic,
            )
            top_levels = [
                constructor(
                    *[TypeVariable.fresh() for _ in range(arity - 1)],
                    AtomicType("output_type"),
                )
                for arity, constructor in constructors_lambda
            ]
            top_levels = [x for x in top_levels if isinstance(x, ArrowType)]
            expanded = []
            for top_level in top_levels:
                expanded += expansions(
                    top_level,
                    types,
                    constructors_lambda,
                    max_expansion_steps=self.max_expansion_steps,
                    max_overall_depth=self.lambda_parameters["max_type_depth"],
                )
            expanded = sorted(expanded, key=str)
            sym_to_productions["<lambda>"] = [
                LambdaProduction(i, LambdaTypeSignature(x.input_type))
                for i, x in enumerate(expanded)
            ]

            variable_types = sorted(
                {
                    input_type
                    for function_type in expanded
                    for input_type in function_type.input_type
                },
                key=str,
            )
            sym_to_productions["<variable>"] = [
                VariableProduction(
                    type_id, VariableTypeSignature(variable_type, index_in_env)
                )
                for type_id, variable_type in enumerate(variable_types)
                for index_in_env in range(self.lambda_parameters["max_env_depth"])
            ]
            # don't prune and reindex variables
            stable_symbols.add("<variable>")

        if self.prune:
            assert self.target_types is not None
            sym_to_productions = prune(
                sym_to_productions,
                self.target_types,
                care_about_variables=False,
                type_depth_limit=self.max_overall_depth,
                env_depth_limit=self.max_env_depth,
                stable_symbols=stable_symbols,
                tolerate_pruning_entire_productions=self.tolerate_pruning_entire_productions,
            )
            if self.prune_variables:
                sym_to_productions = prune(
                    sym_to_productions,
                    self.target_types,
                    care_about_variables=True,
                    type_depth_limit=self.max_overall_depth,
                    env_depth_limit=self.max_env_depth,
                    stable_symbols=stable_symbols,
                    tolerate_pruning_entire_productions=self.tolerate_pruning_entire_productions,
                )
        if "<variable>" in sym_to_productions:
            sym_to_productions["<variable>"] = clean_variables(
                sym_to_productions["<variable>"]
            )
        dsl = make_dsl(
            sym_to_productions,
            copy.copy(self.target_types),
            self.max_overall_depth,
            self.max_env_depth,
        )
        return dsl


def clean_variables(variable_productions):
    type_to_idx = {prod.type_signature().variable_type for prod in variable_productions}
    type_to_idx = {t: i for i, t in enumerate(sorted(type_to_idx, key=str))}
    variable_productions = [
        prod.with_index(type_to_idx[prod.type_signature().variable_type])
        for prod in variable_productions
    ]
    return variable_productions


def make_dsl(sym_to_productions, valid_root_types, max_type_depth, max_env_depth):
    return DSL(
        [prod for prods in sym_to_productions.values() for prod in prods],
        valid_root_types,
        max_type_depth,
        max_env_depth=max_env_depth,
    )


def prune(
    sym_to_productions,
    target_types,
    *,
    care_about_variables,
    type_depth_limit,
    env_depth_limit,
    stable_symbols,
    tolerate_pruning_entire_productions,
):
    dsl = make_dsl(sym_to_productions, target_types, type_depth_limit, env_depth_limit)
    symbols = dsl.constructible_symbols(care_about_variables=care_about_variables)
    new_sym_to_productions = {}
    for original_symbol, prods in sym_to_productions.items():
        new_sym_to_productions[original_symbol] = [
            x for x in prods if x.symbol() in symbols
        ]
        if (
            len(new_sym_to_productions[original_symbol]) == 0
            and not tolerate_pruning_entire_productions
        ):
            raise TypeError(
                f"All productions for {original_symbol} were pruned. "
                f"Check that the target types are correct."
            )
        if original_symbol in stable_symbols:
            continue
        new_sym_to_productions[original_symbol] = Production.reindex(
            new_sym_to_productions[original_symbol]
        )
    return new_sym_to_productions



from typing import Tuple

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def attempt_to_sample_tree_dist(
    dist: TreeDistribution,
    rng: np.random.RandomState,
    depth_limit,
    ancestors: Tuple[Tuple[int, int], ...],
    parent: int,
    preorder_mask: PreorderMask,
) -> SExpression:
    """
    Attempt to sample a program from the distribution, conditioned on the depth limit.

    Args:
        rng: The random number generator to use.
        depth_limit: The maximum depth of the program.
        parents: The parents of the current node, to a limit of dist.limit

    Raises:
        TooDeepError: If the program is too deep.
    """

    if depth_limit < 0:
        raise TooDeepError()
    root_sym, root_arity = dist.symbols[parent]
    children = []
    for i in range(root_arity):
        key = ancestors + ((parent, i),)
        key = key[-dist.limit :]
        possibilites, weights = dist.sampling_dict_arrays[key]
        mask = preorder_mask.compute_mask(i, possibilites)
        possibilites, weights = possibilites[mask], weights[mask]
        if len(possibilites) == 0:
            raise ValueError(f"No valid productions for {key}")
        weights /= weights.sum()
        child_idx = rng.choice(possibilites, p=weights)
        preorder_mask.on_entry(i, child_idx)
        children.append(
            attempt_to_sample_tree_dist(
                dist,
                rng,
                depth_limit - 1,
                ancestors=key,
                parent=child_idx,
                preorder_mask=preorder_mask,
            )
        )
        preorder_mask.on_exit(i, child_idx)
    return SExpression(root_sym, tuple(children))


def sample_tree_dist(
    dist: TreeDistribution, rng: np.random.RandomState, depth_limit
) -> SExpression:
    """
    Sample a program from the distribution, conditioned on the depth limit.

    Args:
        rng: The random number generator to use.
        depth_limit: The maximum depth of the program.
        parents: The parents of the current node, to a limit of dist.limit
    """
    while True:
        try:
            preorder_mask = dist.mask_constructor(dist)
            preorder_mask.on_entry(0, 0)
            return attempt_to_sample_tree_dist(
                dist,
                rng,
                depth_limit,
                ancestors=(),
                parent=0,
                preorder_mask=preorder_mask,
            )
        except TooDeepError:
            continue


class TooDeepError(Exception):
    pass
